#include <getopt.h>
#include <highgui.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

void procesar (char* input_filename, unsigned char *src, unsigned char *dst, int m, int n, int row_size);
const char* nombre_programa;

int main( int argc, char** argv ) {
	// Si se ejecuta sin parametros ni opciones
	if (argc == 1) {
		printf("Falta especificar el nombre del archivo\n");
		exit ( EXIT_SUCCESS );
	}

	// Verifico nombre de archivo
	const char *nomb_arch_entrada = argv[1];
	if (access( nomb_arch_entrada, F_OK ) == -1) {
		printf("Error al intentar abrir el archivo: %s.\n", nomb_arch_entrada);
		exit ( EXIT_SUCCESS );
	}

	IplImage *src = 0;
	IplImage *dst = 0;

	// Cargo la imagen
	if( (src = cvLoadImage (nomb_arch_entrada, CV_LOAD_IMAGE_GRAYSCALE)) == 0 )
		exit(EXIT_FAILURE);

	// Creo una IplImage para cada salida esperada
	if( (dst = cvCreateImage (cvGetSize (src), IPL_DEPTH_8U, 1) ) == 0 )
		exit(EXIT_FAILURE);

	//Proceso la imagen
	procesar(nomb_arch_entrada, (unsigned char*)src->imageData, 
		(unsigned char*)dst->imageData, src->height, src->width, src->widthStep);

	//Libero memoria
	cvReleaseImage(&src);
	cvReleaseImage(&dst);

	return 0;
}

void procesar (char* input_filename, unsigned char *src, unsigned char *dst, int h, int w, int row_size) {
	printf("w = %d, h = %d, rs = %d \n", w, h, row_size);

	w = ((w + 1) >> 1) << 1;
	int act = 0;
	int x;
	int y;

	int minx = w;
	int miny = h;
	int maxx = 0;
	int maxy = 0;

	for (y = 0; y < h; y ++) {
		for (x = 0; x < w; x ++, act ++, src++) {
			char val = (src[0] > 128) ? 1 : 0;
			dst[act] = val;

			minx = (val && x < minx) ? x : minx;
			miny = (val && y < miny) ? y : miny;
			maxx = (val && x > maxx) ? x : maxx;
			maxy = (val && y > maxy) ? y : maxy;
		}
		for (; x < row_size; x ++, src++);
	}

	char filename[256];
	strcpy(filename, input_filename);
        strcpy(filename + strlen(filename), ".out");

	printf("min(X, Y) = (%d, %d)\n", minx, miny);
	printf("max(X, Y) = (%d, %d)\n", maxx, maxy);
	printf("max(X, Y) - min(X, Y) = (%d, %d)\n", maxx - minx, maxy - miny);

	FILE *fh = fopen(filename, "wb");
	if (fh == NULL) {
		printf("Error al abrir 'life.out'!/n");
	} else {
		char buf[50];
		int px = 0;
		int py = 0;
		for (x = 0; x < act; x+=2, px ++) {
			if (x % w == 0) {
				px = 0;
				py ++;
			}

			//unsigned char val = dst[x];
			unsigned char val = (dst[x + 1] << 4) + dst[x];
			if (val != 0) {
				//sprintf(buf, "set_at(state[actState], %d, %d, %d); \n", px - minx, py, val);
				//sprintf(buf, "states[%d + %d * HWID] = %d; \n", px, py, val);
				sprintf(buf, "stateArr[%d][%d] = %d; \n", py, px, val);
				fputs(buf, fh);
			}
		}
		fclose(fh);
	}
	printf("Done\n");
}


