#include "fat16_driver.h"
#include "../kernel/mmu.h"
#include "../kernel/i386.h"
#include "../kernel/sched.h"

//char read_buff[512];
char *cluster_buff;
boot_sector bs;
partition_info primary;

uint root_dir_sectors;
uint first_data_sector;
uint first_fat_sector;
uint data_sectors;
uint total_clusters;
uint root_cluster;

uint bytes_per_cluster;

ushort *fat_table;

fs_directory root;

uchar init_fat16() {
	scr_printf("%C[FS]%C Inicializando el driver del filesystem\n", YELLOW, WHITE);
	
	//Leo el MBR del disco
	if (read_hdd((char*) &bs, 0, 1) == -1) return -1;
	
	//Me interesa la particion 1 nada mas
	memcpy(&primary, &(((char *) &bs)[0x1BE]), sizeof(primary));
	
	//El id de FAT16 es 6 o 4
	if (primary.id != 6 && primary.id != 4) return -1;
	root_dir_sectors = ((bs.root_entry_count * 32) + (bs.bytes_per_sector - 1)) / 
		bs.bytes_per_sector;
	first_data_sector = bs.reserved_sector_count + (bs.table_count * bs.table_size_16);
	first_fat_sector = bs.reserved_sector_count;
	data_sectors = bs.total_sectors_16 - (bs.reserved_sector_count + 
		(bs.table_count * bs.table_size_16) + root_dir_sectors);
	total_clusters = data_sectors / bs.sectors_per_cluster;
	
	root_cluster = first_data_sector;
	
	//Inicializo el directorio root
	root.subdirectories = NULL;
	root.parent = NULL;
	root.files = NULL;
	root.cluster = 0;
	root.name = "ROOT";
	root.next = &root;
	root.prev = &root;
	
	bytes_per_cluster = bs.sectors_per_cluster * 512;

	//NOTA, por estar usando malloc, tras saltar a la 1ra tarea, la info solo es accesible 
	//por el proceso con id 0, o sea, por la tarea idle.
	
	//Cargo la tabla de clusters
	fat_table = (ushort *) malloc((first_data_sector - first_fat_sector) * 512);
	read_hdd((char*) fat_table, first_fat_sector, first_data_sector - first_fat_sector);
	
	cluster_buff = (char *) malloc(bytes_per_cluster);
	
	//Leo el contenido
	int s;
	char *lfn = NULL;
	uint sector = root_cluster;
	for (s = root_dir_sectors; s > 0; s --) {
		if (read_hdd(cluster_buff, sector, 1) == -1) return -1;
		lfn = parse_cluster(cluster_buff, 16, lfn, &root);
	
		sector ++;
	}

	//debug_current_status("Estado de la MMU");
	
	//Parsea el contenido de los subdirectorios
	parse_content(root.subdirectories);
	//scr_printf("\n\n");
	//print_fs(&root, 0);
	
	//unsigned int oeoe = 0;
	//scr_printf("conway/sse.tsk [%d]\n", (int) open_file("conway/sse.tsk", &oeoe));
	//scr_printf("a.txt [%d]\n", (int) open_file("a.txt"));
	//scr_printf("a.txt2 [%d]\n", (int) open_file("a.txt2"));
	//scr_printf("bin/compile.sh [%d]\n", (int) open_file("bin/compile.sh"));
	//scr_printf("bin/ [%d]\n", (int) open_file("bin/"));
	//while(1);
	
	/*
	int i;
	//if (read_hdd(cluster_buff, first_fat_sector, 1) == -1) return -1;
	if (read_cluster(cluster_buff, 3) == -1) return -1;
	scr_printf("\n%C[CLUSTER 3]%C\n", WHITE, YELLOW);
	
	int k;
	for (i = 0; i < 16 * bs.sectors_per_cluster; i++) {
		longname_info *linfo = (longname_info *) &(cluster_buff[i*32]);
		file_info *finfo = (file_info *) &(cluster_buff[i*32]);
		if (cluster_buff[i*32] != 0 && cluster_buff[i*32] != 0xE5) {
			if (linfo->attribute == 0x0F) {
				scr_printf("[LF] - ");
				for (k = 0; k < 5; k++) scr_printf("%c", linfo->char1[k]);
				for (k = 0; k < 6; k++) scr_printf("%c", linfo->char2[k]);
				for (k = 0; k < 2; k++) scr_printf("%c", linfo->char3[k]);
			} else {
				scr_printf("[EN] - ");
				for (k = 0; k < 8; k++) scr_printf("%c", finfo->name[k]);
				scr_printf(".");
				for (k = 0; k < 3; k++) scr_printf("%c", finfo->name[8+k]);
				
				scr_printf("%C cluster=%C%d%C", WHITE, CYAN, (finfo->cluster_h << 16) +
					finfo->cluster_l, YELLOW);
			}
			scr_printf("\n");
		}
	}
	//*/
	/*
	scr_printf("%Croot_dir_sectors = %C%d\n", WHITE, CYAN, root_dir_sectors);
	scr_printf("%Cfirst_data_sector = %C%d\n", WHITE, CYAN, first_data_sector);
	scr_printf("%Cfirst_fat_sector = %C%d\n", WHITE, CYAN, first_fat_sector);
	scr_printf("%Cdata_sectors = %C%d\n", WHITE, CYAN, data_sectors);
	scr_printf("%Ctotal_clusters = %C%d\n", WHITE, CYAN, total_clusters);
	scr_printf("%Csectors_per_cluster = %C%d\n", WHITE, CYAN, bs.sectors_per_cluster);
	
	scr_printf("%CSector %d:", WHITE, root_cluster);
	for (i = 0; i < 512 * bs.sectors_per_cluster; i++) {
		if (i % 40 == 0) {
			scr_printf("\n%C[%d]: %C", WHITE, i, CYAN);
		} else if (i % 20 == 0) {
			scr_printf("%C - [%d]: %C", WHITE, i, CYAN);
		}
		scr_printf(" %h", (uchar) cluster_buff[i]);
	}
	scr_printf("\n");
	//*/
	/*
	scr_printf("%C.bootable=%C%d\n", WHITE, CYAN, primary.bootable);
	scr_printf("%C.start_head=%C%d\n", WHITE, CYAN, primary.start_head);
	scr_printf("%C.start_sector=%C%d\n", WHITE, CYAN, primary.start_sector & 0x3F);
	scr_printf("%C.start_cyl=%C%d\n", WHITE, CYAN, 
		((primary.start_cyl_h << 8) + primary.start_cyl_l));
	scr_printf("%C.id=%C%d\n", WHITE, CYAN, primary.id);
	scr_printf("%C.end_head=%C%d\n", WHITE, CYAN, primary.end_head);
	scr_printf("%C.end_sector=%C%d\n", WHITE, CYAN, primary.end_sector & 0x3F);
	scr_printf("%C.end_cyl=%C%d\n", WHITE, CYAN, 
		((primary.end_cyl_h << 8) + primary.end_cyl_l));
	scr_printf("%C.relative_sector=%C%d\n", WHITE, CYAN, primary.relative_sector);
	scr_printf("%C.total_sectors=%C%d\n", WHITE, CYAN, primary.total_sectors);
	*/
	
	//while (s != 3);
	
	return 0;
}

uchar read_cluster(char *buff, uint cluster) {
	return read_hdd(buff, (cluster - 2) * bs.sectors_per_cluster + 
		first_data_sector + root_dir_sectors, bs.sectors_per_cluster);
}

char *parse_cluster(char *buffer, uint max_entries, char* lfn, fs_directory *parent) {
	uint i, k;
	longname_info *linfo = (longname_info *) buffer;
	file_info *finfo = (file_info *) buffer;
	for (i = 0; i < max_entries; i++) {
		char fstchar = buffer[i*32];
		//Si el primer byte es 0 (disponible) o 0xE5 (borrado), no tomo
		//en cuenta la entrada
		if ((fstchar != 0) && (fstchar != 0xE5)) {
			if (linfo->attribute == 0x0F) {
				//Long file, tengo que sacar el nombre y concatenarlo con el
				//lfn de antes
				uchar cont = 1;
				char name[14];
				uint name_i = 0;
				k = 0;
				while (k < 5 && cont) {
					cont = ((char) linfo->char1[k]) != 0xFF;
					name[name_i ++] = cont ? ((char) linfo->char1[k]) : 0;
					k ++;
				}
				k = 0;
				while (k < 6 && cont) {
					cont = ((char) linfo->char2[k]) != 0xFF;
					name[name_i ++] = cont ? ((char) linfo->char2[k]) : 0;
					k ++;
				}
				k = 0;
				while (k < 2 && cont) {
					cont = ((char) linfo->char3[k]) != 0xFF;
					name[name_i ++] = cont ? ((char) linfo->char3[k]) : 0;
					k ++;
				}
				name[name_i] = 0;
				
				//Concateno con lfn
				if (lfn != NULL) {
					char *olfn = lfn;
					lfn = strcat(name, olfn);
					free(olfn);
				} else {
					lfn = strdup(name);
				}
			} else {
				//Es un archivo/directorio
				if (lfn == NULL) {
					//No tiene nombre largo, lee el del archivo
					
					char name[13];
					uint nospace = 0;
					for (k = 0; k < 8; k++) {
						if (finfo->name[k] != ' ') nospace = k;
						name[k] = finfo->name[k];
					}
					if ((finfo->attributes & 0x10) == 0) {
						name[++nospace] = '.';
						for (k = 8; k < 11; k++) name[++nospace] = finfo->name[k];
					}
					name[++nospace] = 0;
					
					lfn = strdup(name);
				}
				
				if (finfo->attributes & 0x10) {
					if (strcmp(lfn, ".") != 0 && strcmp(lfn, "..") != 0) {
						//Directorio distinto de "." y ".."
						fs_directory *dir = (fs_directory *) malloc(sizeof(fs_directory));
						
						//Veo como lo meto en la lista
						if (parent->subdirectories == NULL) {
							//Es el primer directorio
							dir->prev = dir;
							dir->next = dir;
							parent->subdirectories = dir;
						} else {
							dir->prev = parent->subdirectories->prev;
							dir->next = parent->subdirectories;
							parent->subdirectories->prev->next = dir;
							parent->subdirectories->prev = dir;
						}
						
						dir->cluster = (finfo->cluster_h << 16) + finfo->cluster_l;
						dir->subdirectories = NULL;
						dir->parent = parent;
						dir->files = NULL;
						dir->name = lfn;
					} else {
						free(lfn);
					}
				} else {
					//Archivo
					fs_file *file = (fs_file *) malloc(sizeof(fs_file));
					
					//Veo como lo meto en la lista
					if (parent->files == NULL) {
						//Es el primer archivo
						file->prev = file;
						file->next = file;
						parent->files = file;
					} else {
						file->prev = parent->files->prev;
						file->next = parent->files;
						parent->files->prev->next = file;
						parent->files->prev = file;
					}
					
					file->cluster = (finfo->cluster_h << 16) + finfo->cluster_l;
					file->size = finfo->size;
					file->parent = parent;
					file->name = lfn;
				}
				
				lfn = NULL;
			}
		}
		//*/
		
		linfo ++;
		finfo ++;
	}
	
	return lfn;
}

void parse_content(fs_directory *dir) {
	if (dir == NULL) return;
	fs_directory *fstdir = dir;
	
	do {
		uint act_cluster = dir->cluster;
		
		char *lfn = NULL;
		while (act_cluster != 0 && act_cluster <= 0xFFF7) {
			if (read_cluster(cluster_buff, act_cluster) == -1) return;
			lfn = parse_cluster(cluster_buff, 16 * bs.sectors_per_cluster, lfn, dir);
			
			act_cluster = fat_table[act_cluster - 2];
		}
		
		//Parsea los subdirectorios
		parse_content(dir->subdirectories);
		
		dir = dir->next;
	} while (dir != fstdir);
}


fs_directory *get_directory(const char *path) {
	fs_directory *currDir = &root;
	int e = 0;
	path --;
	
	while (currDir != NULL && e != -1) {
		do {
			path += e + 1;
			e = indexOf(path, '/');
		} while (e == 0);
		if (e != -1) {
			fs_directory *sdir = currDir->subdirectories;
			currDir = NULL;
			if (sdir != NULL) {
				fs_directory *fstdir = sdir;
				do {
					if (strncmp(sdir->name, path, e) == 0) {
						currDir = sdir;
						break;
					}
					sdir = sdir->next;
				} while (sdir != fstdir);
			}
		}
	}
	
	return currDir;
}

int file_exists(const char *path) {
	fs_directory *currDir = get_directory(path);
	if (currDir == NULL) return 0;
	
	int e = 0;
	while (e != -1) {
		e = indexOf(path, '/');
		path += e + 1;
	}
	
	fs_file *file = currDir->files;
	if (file == NULL) return 0;
	
	do {
		if (strcmp(file->name, path) == 0) {
			return 1;
		}
		file = file->next;
	} while (file != currDir->files);
	
	return 0;
}

void *open_file(const char *path, unsigned int *file_size) {
	fs_directory *currDir = get_directory(path);
	if (currDir == NULL) return NULL;
	
	int e = 0;
	while (e != -1) {
		e = indexOf(path, '/');
		path += e + 1;
	}
	
	fs_file *file = currDir->files;
	if (file == NULL) return NULL;

	do {
		if (strcmp(file->name, path) == 0) {
			int reads = ((file->size + bytes_per_cluster - 1) / bytes_per_cluster);
			char *ret = (char *) malloc(reads * bytes_per_cluster);
			int cluster = file->cluster;
			
			char *curr = ret;
			//scr_printf("%CENCONTRADO (%d bytes - r=%d)%C\n", YELLOW, file->size, reads, WHITE);
			while (reads -- > 0) {
				//scr_printf("Leyendo cluster %d [%d bytes]\n", cluster, bytes_per_cluster);
				
				if (read_cluster(curr, cluster) == -1) {
					scr_printf("%CERROR al leer el cluster\n", RED);
					free(ret);
					return NULL;
				}
				curr += bytes_per_cluster;
				cluster = fat_table[cluster];
			}
			
			*file_size = file->size;
			ret[file->size] = 0;
			//scr_printf("%C%d bytes: \n%s\n", WHITE, file->size, ret);
			return ret;
		}
		file = file->next;
	} while (file != currDir->files);
	
	return NULL;
}

void print_fs(fs_directory *parent, uint spaces) {
	if (parent == NULL) return;
	
	scr_printf("%C%p%s\n", WHITE, spaces, parent->name);
	spaces += 4;
	fs_directory *sdir = parent->subdirectories;
	if (sdir != NULL) {
		do {
			print_fs(sdir, spaces);
			sdir = sdir->next;
		} while (sdir != parent->subdirectories);
	}
	
	fs_file *file = parent->files;
	if (file != NULL) {
		scr_printf("%C", CYAN);
		do {
			scr_printf("%p%s (%d bytes)\n", spaces, file->name, file->size);
			file = file->next;
		} while (file != parent->files);
	}
}



int get_directory_contents(const char *path, directory_contents *out) {
	fs_directory *dir = get_directory(path);
	if (dir == NULL) return NULL;

	//Primero cuenta archivos y directorios
	int dirCount = 0;
	fs_directory *d = dir->subdirectories;

	if (d != NULL) {
		do {

			dirCount ++;
			d = d->next;
		} while (d != dir->subdirectories);
	}

	int fileCount = 0;
	fs_file *f = dir->files;
	if (f != NULL) {
		do {
			fileCount ++;
			f = f->next;
		} while (f != dir->files);
	}

	//Prepara memoria para cada los punteros
	out->dirs = (char**) malloc((dirCount+1) * sizeof(char*));
	out->files = (char**) malloc((fileCount+1) * sizeof(char*));


	//Llena la lista de nombres
	int i = 0;

	//d = dir->subdirectories;
	for (i = 0; i < dirCount; i++) {
		out->dirs[i] = strdup(d->name);
		d = d->next;
	}
	out->dirs[dirCount] = 0;

	//f = dir->files;
	for (i = 0; i < fileCount; i++) {
		out->files[i] = strdup(f->name);
		f = f->next;
	}
	out->files[fileCount] = 0;

	return 1;
}

int run_files_as_program(const char *file1, const char *file2) {
	if (!file_exists(file1)) return -1;
	if (!file_exists(file2)) return -2;

	//Cargo el contenido de cada archivo en memoria
	unsigned int size1, size2;
	unsigned int mem1 = (unsigned int) open_file(file1, &size1);
	unsigned int mem2 = (unsigned int) open_file(file2, &size2);

	//Creo los procesos
	crear_proceso(mem1, (size1+4095) / 4096);
	crear_proceso(mem2, (size2+4095) / 4096);

	//Libero el contenido de los archivos en memoria
	free((void *) mem1);
	free((void *) mem2);

	//Cede el resto de su tiempo
	forced_switch_task();

	return 1;
}
int run_file_as_program(const char *file) {
	if (!file_exists(file)) return -1;

	unsigned int size;
	unsigned int mem = (unsigned int) open_file(file, &size);
	crear_proceso(mem, (size+4095) / 4096);
	free((void *) mem);

	//Cede el resto de su tiempo
	forced_switch_task();

	return 1;
}