#ifndef HDD_DRIVER_H
#define FAT16_DRIVER_H

#include "../common/utils.h"
#include "../common/types.h"
#include "../kernel/mmu.h"
#include "hdd_driver.h"

typedef struct fs_directory fs_directory;
typedef struct fs_cluster fs_cluster;
typedef struct fs_file fs_file;

typedef struct partition_info {
	uchar bootable;
	uchar start_head;
	uchar start_sector:6;
	uchar start_cyl_h:2;
	uchar start_cyl_l;
	uchar id;
	uchar end_head;
	uchar end_sector:6;
	uchar end_cyl_h:2;
	uchar end_cyl_l;
	uint relative_sector;
	uint total_sectors;
} __attribute__((__packed__)) partition_info;
 
typedef struct boot_sector {
	unsigned char 		bootjmp[3];
	unsigned char 		oem_name[8];
	unsigned short 	    bytes_per_sector;
	unsigned char		sectors_per_cluster;
	unsigned short		reserved_sector_count;
	unsigned char		table_count;
	unsigned short		root_entry_count;
	unsigned short		total_sectors_16;
	unsigned char		media_type;
	unsigned short		table_size_16;
	unsigned short		sectors_per_track;
	unsigned short		head_side_count;
	unsigned int 		hidden_sector_count;
	unsigned int 		total_sectors_32;
 
	unsigned char		bios_drive_num;
	unsigned char		reserved1;
	unsigned char		boot_signature;
	unsigned int		volume_id;
	unsigned char		volume_label[11];
	unsigned char		fat_type_label[8];
} __attribute__((packed)) boot_sector;

typedef struct file_info {
	char name[11];
	uchar attributes;
	uchar reserved;
	uchar creation_time_tenths;
	ushort creation_time;
	ushort creation_date;
	ushort last_access_date;
	ushort cluster_h;
	ushort last_modification_time;
	ushort last_modification_date;
	ushort cluster_l;
	uint size;
} __attribute__((packed)) file_info;

typedef struct longname_info {
	uchar order;
	ushort char1[5];
	uchar attribute;
	uchar type;
	uchar checksum;
	ushort char2[6];
	ushort reserved;
	ushort char3[2];
} __attribute__((packed)) longname_info;



struct fs_directory {
	fs_directory *subdirectories;
	fs_directory *parent;
	fs_directory *prev;
	fs_directory *next;
	fs_file *files;
	uint cluster;
	char *name;
};

struct fs_file {
	fs_directory *parent;
	fs_file *prev;
	fs_file *next;
	char *name;
	uint cluster;
	uint size;
};

uchar init_fat16();
uchar read_cluster(char *buff, uint cluster);
char *parse_cluster(char *buffer, uint max_entries, char* lfn, fs_directory *parent);
void parse_content(fs_directory *dir);

void print_fs(fs_directory *parent, uint spaces);

int get_directory_contents(const char *path, directory_contents *out);
int run_files_as_program(const char *file1, const char *file2);
int run_file_as_program(const char *file);
void *open_file(const char *path, unsigned int *file_size);

#endif