#include "graphic.h"
#include "../kernel/i386.h"
#include "../kernel/mmu.h"
#include "../common/utils.h"

#define set_letter(chr, l0, l1, l2, l3, l4, l5, l6, l7, l8, l9, lA, lB) \
	letter[chr][0] = l0; \
	letter[chr][1] = l1; \
	letter[chr][2] = l2; \
	letter[chr][3] = l3; \
	letter[chr][4] = l4; \
	letter[chr][5] = l5; \
	letter[chr][6] = l6; \
	letter[chr][7] = l7; \
	letter[chr][8] = l8; \
	letter[chr][9] = l9; \
	letter[chr][10] = lA; \
	letter[chr][11] = lB; \

typedef void (drawByte_fnt) (uchar *, uint, uint, uint);
typedef int (drawStringRGB_fnt) (uchar *, uint, uint, uint);
typedef int (drawStringByte_fnt) (uchar *, uchar, uint, uint);
drawByte_fnt *drawByte_fn = drawByte_8bpp;
drawStringRGB_fnt *drawStringRGB_fn = drawStringRGB_8bpp;
drawStringByte_fnt *drawStringByte_fn = drawStringByte_8bpp;

#define B8(num) 

int start_vesa() {
	int i;
	for (i = 0; i < 256; i++) {
		letter[i][0] = 	0b00000000;
		letter[i][1] = 	0b00111100;
		letter[i][2] = 	0b01000010;
		letter[i][3] = 	0b01000010;
		letter[i][4] = 	0b01000010;
		letter[i][5] = 	0b00000100;
		letter[i][6] = 	0b00001000;
		letter[i][7] = 	0b00001000;
		letter[i][8] = 	0b00001000;
		letter[i][9] = 	0b00000000;
		letter[i][10] = 0b00001000;
		letter[i][11] = 0b00001000;
	}
	
	setLetters();

	actual_mode = 0;
	//Busco la info de VESA
	if (get_vesa_info() == -1) {
		scr_printf("%C[VIDEO]%C VESA graphic mode - Failed!\n", YELLOW, RED);
		return -1;
	}
	
	//Busco el modo con 800x600x8
	int mode = 0; //find_vesa_mode(800, 600, 24);
	if (mode == 0) {
		mode = find_vesa_mode(1024, 768, 8);
		if (mode == 0) {
			scr_printf("%C[VIDEO]%C VESA graphic mode - Failed!\n", YELLOW, RED);
			return -1;
		}
	}
	
	//Cambio a ese modo
	if (set_vesa_mode(mode) == -1) {
		scr_printf("%C[VIDEO]%C VESA graphic mode - Failed!\n", YELLOW, RED);
		return -1;
	}
	
	//Reseteo la posicion de la 'consola'
	scr_setX(0);
	scr_setY(0);
	
	//Indico el modo actual y dejo lista su info 
	//para consultar en el futuro
	actual_mode = mode;
	get_mode_info(mode);
	
	uint bpp = getBitsPerPixel();
	if (bpp == 32) {
		drawByte_fn = drawByte_32bpp;
		drawStringRGB_fn = drawStringRGB_32bpp;
		drawStringByte_fn = drawStringByte_32bpp;
	} else if (bpp == 24) {
		drawByte_fn = drawByte_24bpp;
		drawStringRGB_fn = drawStringRGB_24bpp;
		drawStringByte_fn = drawStringByte_24bpp;
	} else {
		drawByte_fn = drawByte_8bpp;
		drawStringRGB_fn = drawStringRGB_8bpp;
		drawStringByte_fn = drawStringByte_8bpp;
	}
	
	scr_printf("%C[VIDEO]%C VESA graphic mode - ON!\n", YELLOW, GREEN);
	
	return 0;
}

uint getScreenWidth() {
	if (actual_mode == 0) return 320;
	return mode_info.XResolution;
}
uint getScreenHeight() {
	if (actual_mode == 0) return 240;
	return mode_info.YResolution;
}
uint getBytesPerScanLine() {
	if (actual_mode == 0) return getScreenWidth() * (getBitsPerPixel() >> 3);
	return mode_info.BytesPerScanLine;
}
uint getVideoMemAddr() {
	if (actual_mode == 0) return 0xA0000;
	if (mode_info.PhysBasePtr == 0) return 0xA0000;
	return mode_info.PhysBasePtr;
}
uint getBitsPerPixel() {
	if (actual_mode == 0) return 8;
	return mode_info.BitsPerPixel;
}
uint getVideoMemorySize() {
	return (getScreenWidth() * getScreenHeight() * getBitsPerPixel()) >> 3;
}


void drawByte(uchar *data, uint size, uint x, uint y) {
	//uint dir_entry = mapear_memoria_video(rcr3());
	drawByte_fn(data, size, x, y);
	//unmapear_memoria_video(rcr3(), dir_entry);
}
void drawByte_8bpp(uchar *data, uint size, uint x, uint y) {
	uint w = getBytesPerScanLine();
	uchar *video = (uchar*) getVideoMemAddr();
	
	video += y * w + x;
	int i = 0;
	for (; i < size; i++) 
		video[i] = data[i] & 0xF;
}
void drawByte_24bpp(uchar *data, uint size, uint x, uint y) {
	uint w = getBytesPerScanLine();
	uchar *video = (uchar*) getVideoMemAddr();
	
	video += y * w + 3 * x;
	int i = 0;
	int s = 0;
	for (; i < size; i++) {
		video[s++] = data[i];
		video[s++] = data[i];
		video[s++] = data[i];
		//video[4*i + 3] = data[i];
	}
}

void drawByte_32bpp(uchar *data, uint size, uint x, uint y) {
	uint w = getBytesPerScanLine() >> 2;
	uint *video = (uint*) getVideoMemAddr();
	
	video += y * w + x;
	int i = 0;
	for (; i < size; i++) {
		int val = data[i];
		val += (val << 8) + (val << 16) + (val << 24);
		video[i] = val;
	}
}


int drawStringByte(uchar *data, uchar color, uint x, uint y) {
	//uint dir_entry = mapear_memoria_video(rcr3());
	int ret = drawStringByte_fn(data, color, x, y);
	//unmapear_memoria_video(rcr3(), dir_entry);
	
	return ret;
}
int drawStringByte_8bpp(uchar *data, uchar color, uint x, uint y) {
	if (data[0] == 0) return x;

	uchar *video_base = (uchar*) getVideoMemAddr();
	uint w = getBytesPerScanLine();
	uint scrw = getScreenWidth();
	uint scrh = getScreenHeight();

	while (x < 0) {
		x += scrw;
		y -= w * FONT_HEI;
	}
	while (y < 0) y += scrh;
	
	int offy = y * w;
	
	uchar *video_max = video_base + w * scrh;
	uchar *video = video_base;

	video += (offy + (x / scrw) * w);
	x = x % scrw;
	color &= 0xF;
	
	int i = 0;
	int line_v;
	int line_h;
	uchar ldata;
	
	int act_x = x;
	int act_i = 0;
	int act_line_h = 0;
	int next_i = 0;
	int next_line_h = 0;
	
	while (data[i] != 0) {
		for (line_v = 0; line_v < FONT_HEI; line_v ++) {
			x = act_x;
			i = act_i;
			
			if (data[i] != 0) {
				while (video >= video_max) {
					//Scrolleo todo para arriba FONT_HEI pixels
					scroll_upwards(FONT_HEI);
					video -= (FONT_HEI) * w;
				}
			}
			
			while (data[i] != 0 && x < scrw) {
				if (data[i] == 10 && data[i + 1] == 13) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 2;
					next_line_h = 0;
					break;
				} else if (data[i] == 13 || data[i] == 10) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 1;
					next_line_h = 0;
					break;
				} else {
					int max_line = 8;
					if (data[i] == '\t') {
						ldata = 0;
						max_line = 32;
					} else {
						ldata = letter[data[i]][line_v];
					}
					
					line_h = (i == act_i) ? act_line_h : 0;
					ldata = ldata << line_h;
					while (line_h < max_line && x < scrw) {
						video[x++] = (ldata & 0x80) ? color : 0;
						ldata = ldata << 1;
						line_h ++;
					}
					
					if (x >= scrw) {
						next_i = i;
						next_line_h = line_h;
					}
					i++;
				}
			}
			video += w;
		}
		
		if (act_i != next_i || act_line_h != next_line_h) {
			act_x = 0;
			act_i = next_i;
			act_line_h = next_line_h;
			
			i = 0;
		}
	}
	
	int ret = (int) (video - video_base) + x - w * FONT_HEI - offy;
	return ret;
}

int drawStringByte_24bpp(uchar *data, uchar color, uint x, uint y) {
	if (data[0] == 0) return x;

	uchar *video_base = (uchar*) getVideoMemAddr();
	uint w = getBytesPerScanLine();
	uint scrw = getScreenWidth() * 3;
	uint scrh = getScreenHeight();

	while (x < 0) {
		x += scrw;
		y -= w * FONT_HEI;
	}
	while (y < 0) y += scrh;

	int offy = y * w;
	
	uchar *video_max = video_base + w * scrh;
	uchar *video = video_base;
	
	x *= 3;
	video += (offy + (x / scrw) * w);
	x = x % scrw;
	
	int i = 0;
	int line_v;
	int line_h;
	uchar ldata;
	
	int act_x = x;
	int act_i = 0;
	int act_line_h = 0;
	int next_i = 0;
	int next_line_h = 0;
	
	while (data[i] != 0) {
		for (line_v = 0; line_v < FONT_HEI; line_v ++) {
			x = act_x;
			i = act_i;
			
			if (data[i] != 0) {
				while (video >= video_max) {
					//Scrolleo todo para arriba FONT_HEI pixels
					scroll_upwards(FONT_HEI);
					video -= (FONT_HEI) * w;
				}
			}
			
			while (data[i] != 0 && x < scrw) {
				if (data[i] == 10 && data[i + 1] == 13) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 2;
					next_line_h = 0;
					break;
				} else if (data[i] == 13 || data[i] == 10) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 1;
					next_line_h = 0;
					break;
				} else {
					int max_line = 8;
					if (data[i] == '\t') {
						ldata = 0;
						max_line = 32;
					} else {
						ldata = letter[data[i]][line_v];
					}
					
					line_h = (i == act_i) ? act_line_h : 0;
					ldata = ldata << line_h;
					while (line_h < max_line && x < scrw) {
						video[x++] = (ldata & 0x80) ? color : 0;
						video[x++] = (ldata & 0x80) ? color : 0;
						video[x++] = (ldata & 0x80) ? color : 0;
						ldata = ldata << 1;
						line_h ++;
					}
					
					if (x >= scrw) {
						next_i = i;
						next_line_h = line_h;
					}
					i++;
				}
			}
			video += w;
		}
		
		if (act_i != next_i || act_line_h != next_line_h) {
			act_x = 0;
			act_i = next_i;
			act_line_h = next_line_h;
			
			i = 0;
		}
	}
	
	int ret = (int) (video - video_base) + x - w * FONT_HEI - offy;
	return ret / 3;
}
int drawStringByte_32bpp(uchar *data, uchar color, uint x, uint y) {
	return 0;
}
	
int drawStringRGB(uchar *data, uint color, uint x, uint y) {
	//uint dir_entry = mapear_memoria_video(rcr3());
	int ret = drawStringRGB_fn(data, color, x, y);
	//unmapear_memoria_video(rcr3(), dir_entry);
	
	return ret;
}
int drawStringRGB_8bpp(uchar *data, uint color, uint x, uint y) {
	if (data[0] == 0) return x;

	uchar *video_base = (uchar*) getVideoMemAddr();
	uint w = getBytesPerScanLine();
	uint scrw = getScreenWidth();
	uint scrh = getScreenHeight();

	while (x < 0) {
		x += scrw;
		y -= w * FONT_HEI;
	}
	while (y < 0) y += scrh;

	int offy = y * w;
	
	uchar *video_max = video_base + w * scrh;
	uchar *video = video_base;
	
	video += (offy + (x / scrw) * w);
	x = x % scrw;
	
	int i = 0;
	int line_v;
	int line_h;
	uchar ldata;
	
	int act_x = x;
	int act_i = 0;
	int act_line_h = 0;
	int next_i = 0;
	int next_line_h = 0;
	
	uchar r = (color >> 24) & 0xFF;
	uchar g = (color >> 16) & 0xFF;
	uchar b = (color >> 8) & 0xFF;
	color = (((r+g+b) > 0x17F) ? 8 : 0) | 
		((r > 0) ? 4 : 0) | ((g > 0) ? 2: 0) | ((b > 0) ? 1 : 0);
	
	while (data[i] != 0) {
		for (line_v = 0; line_v < FONT_HEI; line_v ++) {
			x = act_x;
			i = act_i;
			
			if (data[i] != 0) {
				while (video >= video_max) {
					//Scrolleo todo para arriba FONT_HEI pixels
					scroll_upwards(FONT_HEI);
					video -= (FONT_HEI) * w;
				}
			}
			
			while (data[i] != 0 && x < scrw) {
				if (data[i] == 10 && data[i + 1] == 13) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 2;
					next_line_h = 0;
					break;
				} else if (data[i] == 13 || data[i] == 10) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 1;
					next_line_h = 0;
					break;
				} else {
					int max_line = 8;
					if (data[i] == '\t') {
						ldata = 0;
						max_line = 32;
					} else {
						ldata = letter[data[i]][line_v];
					}
					
					line_h = (i == act_i) ? act_line_h : 0;
					ldata = ldata << line_h;
					while (line_h < max_line && x < scrw) {
						video[x++] = (ldata & 0x80) ? color : 0;
						ldata = ldata << 1;
						line_h ++;
					}
					
					if (x >= scrw) {
						next_i = i;
						next_line_h = line_h;
					}
					i++;
				}
			}
			video += w;
		}
		
		if (act_i != next_i || act_line_h != next_line_h) {
			act_x = 0;
			act_i = next_i;
			act_line_h = next_line_h;
			
			i = 0;
		}
	}
	
	int ret = (int) (video - video_base) + x - w * FONT_HEI - offy;
	return ret;
}
int drawStringRGB_24bpp(uchar *data, uint color, uint x, uint y) {
	if (data[0] == 0) return x;

	uchar *video_base = (uchar*) getVideoMemAddr();
	uint w = getBytesPerScanLine();
	uint scrw = getScreenWidth() * 3;
	uint scrh = getScreenHeight();

	while (x < 0) {
		x += scrw;
		y -= w * FONT_HEI;
	}
	while (y < 0) y += scrh;

	int offy = y * w;
	
	uchar *video_max = video_base + w * scrh;
	uchar *video = video_base;
	
	x *= 3;
	video += (offy + (x / scrw) * w);
	x = x % scrw;
	
	int i = 0;
	int line_v;
	int line_h;
	uchar ldata;
	
	int act_x = x;
	int act_i = 0;
	int act_line_h = 0;
	int next_i = 0;
	int next_line_h = 0;
	
	uchar r = (color >> 24) & 0xFF;
	uchar g = (color >> 16) & 0xFF;
	uchar b = (color >> 8) & 0xFF;
	
	while (data[i] != 0) {
		for (line_v = 0; line_v < FONT_HEI; line_v ++) {
			x = act_x;
			i = act_i;
			
			if (data[i] != 0) {
				while (video >= video_max) {
					//Scrolleo todo para arriba FONT_HEI pixels
					scroll_upwards(FONT_HEI);
					video -= (FONT_HEI) * w;
				}
			}
			
			while (data[i] != 0 && x < scrw) {
				if (data[i] == 10 && data[i + 1] == 13) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 2;
					next_line_h = 0;
					break;
				} else if (data[i] == 13 || data[i] == 10) {
					while (x < scrw) 	video[x ++] = 0;
					
					next_i = i + 1;
					next_line_h = 0;
					break;
				} else {
					int max_line = 8;
					if (data[i] == '\t') {
						ldata = 0;
						max_line = 32;
					} else {
						ldata = letter[data[i]][line_v];
					}
					line_h = (i == act_i) ? act_line_h : 0;
					ldata = ldata << line_h;
					while (line_h < max_line && x < scrw) {
						video[x++] = (ldata & 0x80) ? b : 0;
						video[x++] = (ldata & 0x80) ? g : 0;
						video[x++] = (ldata & 0x80) ? r : 0;
						ldata = ldata << 1;
						line_h ++;
					}
					
					if (x >= scrw) {
						next_i = i;
						next_line_h = line_h;
					}
					i++;
				}
			}
			video += w;
		}
		
		if (act_i != next_i || act_line_h != next_line_h) {
			act_x = 0;
			act_i = next_i;
			act_line_h = next_line_h;
			
			i = 0;
		}
	}
	
	int ret = (int) (video - video_base) + x - w * FONT_HEI - offy;
	return ret / 3;
}
int drawStringRGB_32bpp(uchar *data, uint color, uint x, uint y) {
	return 0;
}

void scroll_upwards(uint pixels) {
	uint scrh = getScreenHeight();
	uint w = getBytesPerScanLine();
	uchar *video_old = (uchar*) getVideoMemAddr();
	uchar *video_new = (uchar*) (((uint) video_old) + w * pixels);
	
	uint h = scrh - pixels;
	
	int y, i;
	int x = 0;
	for (y = 0; y < h; y++) {
		for (i = 0; i < w; x++, i++) {
			video_old[x] = video_new[x];
		}
	}
	
	for (y = 0; y < pixels; y++) {
		for (i = 0; i < w; x++, i++) {
			video_old[x] = 0x00;
		}
	}
}


void setLetters() {
	int i;
	
	//Mayusculas
	set_letter('A',
		0b00000000,
		
		0b00010000,
		0b00101000,
		0b00101000,
		0b01000100,
		0b01000100,
		0b01000100,
		0b11111110,
		0b10000010,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('B',
		0b00000000,
		
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000010,
		0b11111100,
		0b10000010,
		0b10000010,
		0b10000100,
		0b11111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('C',
		0b00000000,
		
		0b00111110,
		0b01000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b01000000,
		0b00111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('D',
		0b00000000,
		
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000100,
		0b11111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('E',
		0b00000000,
		
		0b11111110,
		0b10000000,
		0b10000000,
		0b10000000,
		0b11110000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b11111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('F',
		0b00000000,
		
		0b11111110,
		0b10000000,
		0b10000000,
		0b10000000,
		0b11110000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('G',
		0b00000000,
		
		0b00111110,
		0b01000000,
		0b10000000,
		0b10000000,
		0b10011100,
		0b10000010,
		0b10000010,
		0b01000010,
		0b00111100,
		
		0b00000000,
		0b00000000);
		
	set_letter('H',
		0b00000000,
		
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b11111110,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('I',
		0b00000000,
		
		0b11111110,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b11111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('J',
		0b00000000,
		
		0b00011110,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b10000100,
		0b10000100,
		0b01000100,
		0b00111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('K',
		0b00000000,
		
		0b10000100,
		0b10001000,
		0b10010000,
		0b10100000,
		0b11000000,
		0b10100000,
		0b10010000,
		0b10001000,
		0b10000100,
		
		0b00000000,
		0b00000000);
		
	set_letter('L',
		0b00000000,
		
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b11111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('M',
		0b00000000,
		
		0b10000010,
		0b11000110,
		0b10101010,
		0b10010010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('N',
		0b00000000,
		
		0b10000010,
		0b11000010,
		0b10100010,
		0b10100010,
		0b10010010,
		0b10001010,
		0b10001010,
		0b10000110,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('O',
		0b00000000,
		
		0b00111000,
		0b01000100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01000100,
		0b00111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('P',
		0b00000000,
		
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000100,
		0b11111000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b10000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('Q',
		0b00000000,
		
		0b00111000,
		0b01000100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10001010,
		0b01000100,
		0b00111100,
		
		0b00000110,
		0b00001110);
		
	set_letter('R',
		0b00000000,
		
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000100,
		0b11111000,
		0b10001000,
		0b10000100,
		0b10000010,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('S',
		0b00000000,
		
		0b00111110,
		0b01000000,
		0b10000000,
		0b01000000,
		0b00111000,
		0b00000100,
		0b00000010,
		0b00000100,
		0b11111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('T',
		0b00000000,
		
		0b11111110,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('U',
		0b00000000,
		
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01000100,
		0b00111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('V',
		0b00000000,
		
		0b10000010,
		0b10000010,
		0b01000100,
		0b01000100,
		0b01000100,
		0b00101000,
		0b00101000,
		0b00101000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('W',
		0b00000000,
		
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10101010,
		0b10111010,
		0b10111010,
		0b01010100,
		
		0b00000000,
		0b00000000);
		
	set_letter('X',
		0b00000000,
		
		0b10000010,
		0b10000010,
		0b01000100,
		0b00101000,
		0b00010000,
		0b00101000,
		0b01000100,
		0b10000010,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('Y',
		0b00000000,
		
		0b10000010,
		0b01000100,
		0b00101000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('Z',
		0b00000000,
		
		0b11111110,
		0b00000010,
		0b00000100,
		0b00001000,
		0b00010000,
		0b00100000,
		0b01000000,
		0b10000000,
		0b11111110,
		
		0b00000000,
		0b00000000);
	
	//*
	
	
	//Minusculas
	set_letter('a',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b01111100,
		0b00000010,
		0b00000010,
		0b01111110,
		0b10000010,
		0b10000010,
		0b01111100,
		
		0b00000000,
		0b00000000);
		
	set_letter('b',
		0b00000000,
		
		0b10000000,
		0b10000000,
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000100,
		0b11111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('c',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00111110,
		0b01000000,
		0b10000000,
		0b10000000,
		0b10000000,
		0b01000000,
		0b00111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('d',
		0b00000000,
		
		0b00000010,
		0b00000010,
		0b00111110,
		0b01000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01000010,
		0b00111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('e',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00111000,
		0b01000100,
		0b10000010,
		0b11111110,
		0b10000000,
		0b01000000,
		0b00111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('f',
		0b00000000,
		
		0b00001110,
		0b00010000,
		0b11111110,
		0b00100000,
		0b00100000,
		0b00100000,
		0b00100000,
		0b00100000,
		0b00100000,
		
		0b00000000,
		0b00000000);
		
	set_letter('g',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00111110,
		0b01000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01000010,
		0b00111110,
		
		0b00000010,
		0b00111100);
		
	set_letter('h',
		0b00000000,
		
		0b10000000,
		0b10000000,
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('i',
		0b00000000,
		
		0b0000000,
		0b00010000,
		0b00000000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b01111100,
		
		0b00000000,
		0b00000000);
		
	set_letter('j',
		0b00000000,
		
		0b00000100,
		0b00000000,
		0b01111100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		
		0b00000100,
		0b01111000);
		
	set_letter('k',
		0b00000000,
		
		0b10000000,
		0b10000000,
		0b10000100,
		0b10001000,
		0b10010000,
		0b11100000,
		0b10100000,
		0b10010000,
		0b10001000,
		
		0b00000000,
		0b00000000);
		
	set_letter('l',
		0b00000000,
		
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b01111100,
		
		0b00000000,
		0b00000000);
		
	set_letter('m',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b01101100,
		0b10010010,
		0b10010010,
		0b10010010,
		0b10010010,
		0b10010010,
		0b10010010,
		
		0b00000000,
		0b00000000);
		
	set_letter('n',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('o',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00111000,
		0b01000100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01000100,
		0b00111000,
		
		0b00000000,
		0b00000000);
		
	set_letter('p',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b11111000,
		0b10000100,
		0b10000010,
		0b10000010,
		0b10000100,
		0b11111000,
		0b10000000,
		
		0b10000000,
		0b10000000);
		
	set_letter('q',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00111110,
		0b01000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01000010,
		0b00111110,
		
		0b00000010,
		0b00000110);
		
	set_letter('r',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b10111100,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('s',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b01111110,
		0b10000000,
		0b01000000,
		0b00111000,
		0b00000100,
		0b00000010,
		0b11111100,
		
		0b00000000,
		0b00000000);
		
	set_letter('t',
		0b00000000,
		
		0b00000000,
		0b00100000,
		0b11111100,
		0b00100000,
		0b00100000,
		0b00100000,
		0b00100000,
		0b00100000,
		0b00100000,
		
		0b00100100,
		0b00011000);
		
	set_letter('u',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b10000100,
		0b10000100,
		0b10000100,
		0b10000100,
		0b10000100,
		0b01000100,
		0b00111010,
		
		0b00000000,
		0b00000000);
		
	set_letter('v',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b10000010,
		0b10000010,
		0b01000100,
		0b01000100,
		0b00101000,
		0b00101000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('w',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10010010,
		0b01010100,
		0b00101000,
		
		0b00000000,
		0b00000000);
		
	set_letter('x',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b10000010,
		0b01000100,
		0b00101000,
		0b00010000,
		0b00101000,
		0b01000100,
		0b10000010,
		
		0b00000000,
		0b00000000);
		
	set_letter('y',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b10000010,
		0b10000010,
		0b01000100,
		0b01000100,
		0b00101000,
		0b00101000,
		0b00010000,
		
		0b00010000,
		0b00100000);
		
	set_letter('z',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b11111110,
		0b00000100,
		0b00001000,
		0b00010000,
		0b00100000,
		0b01000000,
		0b11111110,
		
		0b00000000,
		0b00000000);
		
	
	
	
	
	//Numeros
	set_letter('0',
		0b00000000,
		
		0b01111100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01111100,
		
		0b00000000,
		0b00000000);
	
	set_letter('1',
		0b00000000,
		
		0b00010000,
		0b00110000,
		0b01010000,
		0b10010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b11111110,
		
		0b00000000,
		0b00000000);
	
	set_letter('2',
		0b00000000,
		
		0b00111000,
		0b01000100,
		0b10000010,
		0b00000100,
		0b00001000,
		0b00010000,
		0b00100000,
		0b01000000,
		0b11111110,
		
		0b00000000,
		0b00000000);
	
	set_letter('3',
		0b00000000,
		
		0b11111000,
		0b00000100,
		0b00000010,
		0b00000010,
		0b01111100,
		0b00000010,
		0b00000010,
		0b00000100,
		0b11111000,
		
		0b00000000,
		0b00000000);
	
	set_letter('4',
		0b00000000,
		
		0b00001000,
		0b00011000,
		0b00101000,
		0b01001000,
		0b10001000,
		0b11111110,
		0b00001000,
		0b00001000,
		0b00001000,
		
		0b00000000,
		0b00000000);
	
	set_letter('5',
		0b00000000,
		
		0b11111110,
		0b10000000,
		0b10000000,
		0b11111000,
		0b00000100,
		0b00000010,
		0b00000010,
		0b10000100,
		0b01111000,
		
		0b00000000,
		0b00000000);
	
	set_letter('6',
		0b00000000,
		
		0b00111110,
		0b01000000,
		0b10000000,
		0b10000000,
		0b10111110,
		0b11000010,
		0b10000010,
		0b10000010,
		0b01111100,
		
		0b00000000,
		0b00000000);
		
	set_letter('7',
		0b00000000,
		
		0b11111110,
		0b00000010,
		0b00000010,
		0b00000100,
		0b00000100,
		0b00001000,
		0b00001000,
		0b00010000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('8',
		0b00000000,
		
		0b01111100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01111100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01111100,
		
		0b00000000,
		0b00000000);
		
	set_letter('9',
		0b00000000,
		
		0b01111100,
		0b10000010,
		0b10000010,
		0b10000010,
		0b10000010,
		0b01111110,
		0b00000010,
		0b00000010,
		0b11111100,
		
		0b00000000,
		0b00000000);
	
	
	
	//Caracteres especiales
	for (i = 0; i < FONT_HEI; i++) letter[' '][i] = 0;
	
	set_letter('!',
		0b00000000,
		
		0b00010000,
		0b00111000,
		0b00111000,
		0b00111000,
		0b00111000,
		0b00111000,
		0b00111000,
		0b00010000,
		0b00000000,
		
		0b00010000,
		0b00010000);
		
	set_letter('@',
		0b00000000,
		
		0b01111100,
		0b10000010,
		0b10000010,
		0b10111010,
		0b10111010,
		0b10100010,
		0b10111100,
		0b10000000,
		0b01111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('#',
		0b00000000,
		
		0b00100010,
		0b00100010,
		0b00100010,
		0b11111110,
		0b01000100,
		0b01000100,
		0b11111110,
		0b10001000,
		0b10001000,
		
		0b00000000,
		0b00000000);
		
	set_letter('$',
		0b00000000,
		
		0b00010000,
		0b01111100,
		0b10010010,
		0b01010000,
		0b00111000,
		0b00010100,
		0b10010010,
		0b01111100,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('%',
		0b00000000,
		
		0b00000010,
		0b11000010,
		0b11000100,
		0b00001000,
		0b00010000,
		0b00100000,
		0b01000110,
		0b10000110,
		0b10000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('&',
		0b00000000,
		
		0b00010000,
		0b00101000,
		0b00010000,
		0b00101010,
		0b01000100,
		0b10001000,
		0b10001000,
		0b01000100,
		0b00111010,
		
		0b00000000,
		0b00000000);
		
	set_letter('*',
		0b00000000,
		
		0b00000000,
		0b00010000,
		0b00010000,
		0b01111100,
		0b00101000,
		0b01000100,
		0b00000000,
		0b00000000,
		0b00000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('(',
		0b00000000,
		
		0b00010000,
		0b00010000,
		0b00100000,
		0b00100000,
		0b01000000,
		0b00100000,
		0b00100000,
		0b00010000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter(')',
		0b00000000,
		
		0b00010000,
		0b00010000,
		0b00001000,
		0b00001000,
		0b00000100,
		0b00001000,
		0b00001000,
		0b00010000,
		0b00010000,
		
		0b00000000,
		0b00000000);
	set_letter('[',
		0b00000000,
		
		0b01110000,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01000000,
		0b01110000,
		
		0b00000000,
		0b00000000);
		
	set_letter(']',
		0b00000000,
		
		0b00011100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00000100,
		0b00011100,
		
		0b00000000,
		0b00000000);
	set_letter('{',
		0b00000000,
		
		0b00010000,
		0b00100000,
		0b00100000,
		0b00100000,
		0b01000000,
		0b00100000,
		0b00100000,
		0b00100000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('}',
		0b00000000,
		
		0b00010000,
		0b00001000,
		0b00001000,
		0b00001000,
		0b00000100,
		0b00001000,
		0b00001000,
		0b00001000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('<',
		0b00000000,
		
		0b00001000,
		0b00010000,
		0b00100000,
		0b01000000,
		0b10000000,
		0b01000000,
		0b00100000,
		0b00010000,
		0b00001000,
		
		0b00000000,
		0b00000000);
		
	set_letter('>',
		0b00000000,
		
		0b00100000,
		0b00010000,
		0b00001000,
		0b00000100,
		0b00000010,
		0b00000100,
		0b00001000,
		0b00010000,
		0b00100000,
		
		0b00000000,
		0b00000000);
		
	set_letter('\'',
		0b00000000,
		
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('"',
		0b00000000,
		
		0b11101110,
		0b11101100,
		0b11001100,
		0b10001000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('/',
		0b00000000,
		
		0b00000100,
		0b00000100,
		0b00001000,
		0b00001000,
		0b00010000,
		0b00100000,
		0b00100000,
		0b01000000,
		0b01000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('\\',
		0b00000000,
		
		0b01000000,
		0b01000000,
		0b00100000,
		0b00100000,
		0b00010000,
		0b00001000,
		0b00001000,
		0b00000100,
		0b00000100,
		
		0b00000000,
		0b00000000);
		
	set_letter('_',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b11111110,
		
		0b00000000,
		0b00000000);
		
	set_letter('|',
		0b00000000,
		
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('-',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b11111110,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('+',
		0b00000000,
		
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		0b11111110,
		0b00010000,
		0b00010000,
		0b00010000,
		0b00010000,
		
		0b00000000,
		0b00000000);
		
	set_letter('=',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b11111110,
		0b00000000,
		0b00000000,
		0b00000000,
		0b11111110,
		0b00000000,
		0b00000000,
		
		0b00000000,
		0b00000000);
		
	set_letter('?',
		0b00000000,
		
		0b00111000,
		0b01000100,
		0b10000010,
		0b10000010,
		0b00000100,
		0b00011000,
		0b00010000,
		0b00010000,
		0b00000000,
		
		0b00010000,
		0b00010000);
		
	set_letter('.',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00110000,
		0b00110000,
		
		0b00000000,
		0b00000000);
		
	set_letter(',',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00110000,
		0b00110000,
		
		0b01100000,
		0b11000000);
		
	set_letter(':',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00110000,
		0b00110000,
		0b00000000,
		0b00000000,
		0b00000000,
		0b00110000,
		0b00110000,
		
		0b00000000,
		0b00000000);
		
	set_letter(';',
		0b00000000,
		
		0b00000000,
		0b00000000,
		0b00000000,
		0b00110000,
		0b00110000,
		0b00000000,
		0b00000000,
		0b00110000,
		0b00110000,
		
		0b01100000,
		0b00000000);
	//*/
}