#ifndef GRAPHIC_H
#define GRAPHIC_H

#include "../common/types.h"

#define FONT_HEI 12

typedef struct VESA_INFO {
	uchar  VESASignature[4];
	unsigned short VESAVersion;
	unsigned long  OEMStringPtr;
	uchar  Capabilities[4];
	unsigned long  VideoModePtr;
	unsigned short TotalMemory;
	unsigned short OemSoftwareRev;
	unsigned long  OemVendorNamePtr;
	unsigned long  OemProductNamePtr;
	unsigned long  OemProductRevPtr;
	uchar  Reserved[222];
	uchar  OemData[256];
} __attribute__ ((packed)) VESA_INFO;

typedef struct MODE_INFO {
	unsigned short ModeAttributes;
	uchar  WinAAttributes;
	uchar  WinBAttributes;
	unsigned short WinGranularity;
	unsigned short WinSize;
	unsigned short WinASegment;
	unsigned short WinBSegment;
	unsigned long  WinFuncPtr;
	unsigned short BytesPerScanLine;
	unsigned short XResolution;
	unsigned short YResolution;
	uchar  XCharSize;
	uchar  YCharSize;
	uchar  NumberOfPlanes;
	uchar  BitsPerPixel;
	uchar  NumberOfBanks;
	uchar  MemoryModel;
	uchar  BankSize;
	uchar  NumberOfImagePages;
	uchar  Reserved_page;
	uchar  RedMaskSize;
	uchar  RedMaskPos;
	uchar  GreenMaskSize;
	uchar  GreenMaskPos;
	uchar  BlueMaskSize;
	uchar  BlueMaskPos;
	uchar  ReservedMaskSize;
	uchar  ReservedMaskPos;
	uchar  DirectColorModeInfo;
	unsigned long  PhysBasePtr;
	unsigned long  OffScreenMemOffset;
	unsigned short OffScreenMemSize;
	uchar  Reserved[206];
} __attribute__ ((packed)) MODE_INFO;


MODE_INFO mode_info;
VESA_INFO vesa_info;

int get_vesa_info();
int get_mode_info(uint mode);
int set_vesa_mode(uint mode);
int find_vesa_mode(uint wid, uint hei, uint dep);
int start_vesa();

uint getScreenWidth();
uint getScreenHeight();
uint getBitsPerPixel();
uint getBytesPerScanLine();
uint getVideoMemAddr();
uint getVideoMemorySize();

int drawStringByte(uchar *data, uchar color, uint x, uint y);
int drawStringByte_8bpp(uchar *data, uchar color, uint x, uint y);
int drawStringByte_24bpp(uchar *data, uchar color, uint x, uint y);
int drawStringByte_32bpp(uchar *data, uchar color, uint x, uint y);
	
int drawStringRGB(uchar *data, uint color, uint x, uint y);
int drawStringRGB_8bpp(uchar *data, uint color, uint x, uint y);
int drawStringRGB_24bpp(uchar *data, uint color, uint x, uint y);
int drawStringRGB_32bpp(uchar *data, uint color, uint x, uint y);

void drawByte(uchar *data, uint size, uint x, uint y);
void drawByte_8bpp(uchar *data, uint size, uint x, uint y);
void drawByte_24bpp(uchar *data, uint size, uint x, uint y);
void drawByte_32bpp(uchar *data, uint size, uint x, uint y);

void scroll_upwards(uint pixels);

void setLetters();

uint actual_mode;
uchar letter[256][FONT_HEI];

#endif
