#include "hdd_driver.h"
#include "../common/utils.h"

uchar hdd_on = 0;
uchar check_hdd() {
	//Verifico si el bus esta 'flotando'
	if (inb(IDE_PORT_STATUS) == 0xFF) return -1;
	
	//Selecciono el dispositivo
	int i = 0;
	outb(IDE_PORT_DRIVE, 0xA0);
	
	//Espero 400ms
	for (; i < 5; i++) 	inb(IDE_PORT_STATUS);
	
	outb(0x3F6, 2);
	
	//Mando IDENTIFY
	outb(IDE_PORT_LBA_L, 0x00);
	outb(IDE_PORT_LBA_M, 0x00);
	outb(IDE_PORT_LBA_H, 0x00);
	outb(IDE_PORT_COMMAND, 0xEC);
	
	uchar status;
	for (; i < 5; i++) 	status = inb(IDE_PORT_STATUS);
	
	//Espero hasta que el flag BUSY se ponga en 0
	while (status & IDE_STATUS_BUSY) status = inb(IDE_PORT_STATUS);
	
	//Espero hasta que el flag DRQ o ERR se ponga en 1
	i = 0;
	while (!(status & (IDE_STATUS_DRQ | IDE_STATUS_ERR)) && i < 100) {
		status = inb(IDE_PORT_STATUS);
		i ++;
	}
	
	if (status & IDE_STATUS_ERR) return -1;
	if (!(status & IDE_STATUS_DRQ)) return -1;
	
	i = 0;
	while (i < 256) {
		inw(IDE_PORT_DATA);
		i ++;
	}
	
	hdd_on = 1;
	scr_printf("%C[HDD]%C HDD driver ready!\n", YELLOW, GREEN);
	
	/*
	char secz[512];
	int j;
	char smt = 1;
	for (j = 80; j < 1024 && smt == 1; j++) {
		for (i = 0; i < 512; i++) secz[i] = 0;
		
		if (read_hdd(secz, j, 1) == -1) {
			scr_printf("Todo mal");
		}
		
		scr_xyprintf(0, 36, YELLOW, "%C[HDD]%C Sector %d", YELLOW, WHITE, j); //BPB[0-3] = %d %d %d\n", YELLOW, WHITE, secz[0], secz[1], secz[2]);
		for (i = 0; i < 512; i++) {
			if (secz[i] != 0) smt = 0;
			if (smt == 0) {
				if (i % 20 == 0) {
					scr_printf("\n%C%d/%h - %C", CYAN, i, i, WHITE);
				} else if (i % 10 == 0) {
					scr_printf(" -- %C%d/%h - %C", CYAN, i, i, WHITE);
				}
				scr_printf("%h ", (uchar) secz[i]);
			}
		}
		if (smt == 0) j= 1500;
		scr_printf("\n");
	}
	//asm("hlt");
	//*/
	return 0;
}

uchar read_hdd(char* buffer, uint lba, uchar sectors) {
	if (hdd_on == 0) return -1;
	
	//Aviso que voy a leer y le paso los 4 bits mas significativos del LBA28
	outb(IDE_PORT_DRIVE, 0xE0 | ((lba >> 24) & 0x0F));
	
	//Pierde algo de tiempo
	outb(IDE_PORT_ERROR, 0x00);
	
	//Paso la cantidad de sectores a leer
	outb(IDE_PORT_SECTORCOUNT, sectors);
	
	//Mando la direccion LBA
	outb(IDE_PORT_LBA_L, (uchar) lba);
	outb(IDE_PORT_LBA_M, (uchar) (lba >> 8));
	outb(IDE_PORT_LBA_H, (uchar) (lba >> 16));
	
	//Read those sectors, you must
	outb(IDE_PORT_COMMAND, 0x20);
	
	//It's dangerous to go alone, wait a bit!
	uchar status;
	int i;
	
	while (sectors > 0) {
		//Espera 400 ns y lee el registro de estado
		for (i = 0; i < 5; i++) 	status = inb(IDE_PORT_STATUS);
		
		//Loopea hasta que el flag BUSY se ponga en 0
		while ((status & IDE_STATUS_BUSY) == IDE_STATUS_BUSY)
			status = inb(IDE_PORT_STATUS);
			
		if (status & IDE_STATUS_DRQ) {
			//OK!
			//Lee 256 words
			asm("rep insw" : : "c"(256), "d"(IDE_PORT_DATA), "D"(buffer));
			buffer += 512;
			
			//Decrementa la cantidad de sectores a leer
			sectors --;
		} else {
			//NNGH!
			return -1;
		}
	}
	return 0;
}
