#include "keyboard.h"
#include "../kernel/sched.h"
#include "../common/utils.h"

#define BREAK_CODE 2
#define SKIP_CODE 2
#define SPECIAL_CODE 1;
#define NORMAL_CODE 0;

char fromScanCode[] = {
	 0,		27,		'1',	'2',	'3',	'4',	'5',	'6',	'7',	'8',	
	'9',	'0',	'-',	']',	 8,	 	 0,		'Q',	'W',	'E'	,	'R',
	'T',	'Y',	'U',	'I',	'O',	'P',	'{',	'}',	'\n',	 0,	
	'A',	'S',	'D',	'F',	'G',	'H',	'J',	'K',	'L',	';',
	39,		168,	 0,		'|',	'Z',	'X',	'C',	'V', 	'B',	'N',
	'M',	',',	'.',	'/',	 0,		'*',	 0,	 	' ',	 0,		 0,
	 0,		 0,		 0,		 0,		 0,		 0,		 0,		 0,		 0,		'7',
	'8', 	'7',	'8',	'9',	'-',	'4',	'5',	'6',	'+',	'1',
	'2',	'3',	'0',	'.'
};
char keysPressed[256];

char shiftPressed = 0;
char buffering = 0;
char lastCode = 0;
char lastChar = 0;

//Si el buffer sube a 8192, creo
//que deja de andar por un tema de espacio
//y habria que ver el makefile
unsigned char k_buffer[1024];
int k_nextReadPos = 0;
int k_nextSavePos = 0;

unsigned char tolowercase(unsigned char l) {
	if (l < 'A' || l > 'Z') return l;
	return 'a' + (l - 'A');
}
unsigned char touppercase(unsigned char l) {
	switch (l) {
		case '0': return ')';
		case '1': return '!';
		case '2': return '@';
		case '3': return '#';
		case '4': return '$';
		case '5': return '%';
		case '6': return '^';
		case '7': return '&';
		case '8': return '*';
		case '9': return '(';

		case '{': return '[';
		case '}': return ']';
		case ';': return ':';
		case '\'': return '"';
		case ',': return '<';
		case '.': return '>';
		case '/': return '?';
		case '\\': return '|';
	}

	return l;
}
unsigned char getCharFromCode(unsigned char code) {
	if (code < 84) {
		return fromScanCode[(int) code];
	}
	return 0;
}

unsigned char processCode(unsigned char code) {
	//Key released or already pressed
	if (code & 0x80) {
		code = code & 0x7F;
		keysPressed[(int) code] = 0;

		if (!keysPressed[0x2A] && !keysPressed[0x36]) 
			shiftPressed = 0;

		return SKIP_CODE;
	}
	if (keysPressed[(int) code]) return SKIP_CODE;

	keysPressed[(int) code] = 1;
	lastCode = code;

	if (code == 0x2A || code == 0x36) {
		//Shift
		shiftPressed = 1;
		return SPECIAL_CODE;
	} else if (code == 0x01) {
		//Escape
		return SPECIAL_CODE;
	}

	return NORMAL_CODE;
}
void processNextCode() {
	lastCode = 0;
	lastChar = 0;
	if (k_nextSavePos == k_nextReadPos) return;

	char code = k_buffer[k_nextReadPos ++];
	if (k_nextReadPos == 1024) k_nextReadPos = 0;

	if (processCode(code) == SKIP_CODE) {
		processNextCode();
	} else {
		if (code != 0) {
			//Setea el lastChar
			//Si no tiene apretado shift, nada mas cambia los caracteres a minuscula
			lastChar = getCharFromCode(code);
			if (shiftPressed == 0) {
				lastChar = tolowercase(lastChar);
			} else {
				lastChar = touppercase(lastChar);
			}
		}
	}
}


void k_initialize() {
	//k_nextReadPos = 0;
	//k_nextSavePos = 0;

	int i = 0;
	for (; i < 256; i++) keysPressed[i] = 0;
}

int special_key_pressed = 0;
void k_gotInput(unsigned char keyCode) {
	if (keyCode == 0x3E && special_key_pressed == 0) {
		//F4 fue presionada
		s_specialKeyPressed();
		special_key_pressed = 1;
	}
	if (keyCode == (0x3E|0x80) && special_key_pressed == 1) {
		//F4 fue presionada
		special_key_pressed = 0;
	}

	if (buffering == 1) {
		k_buffer[k_nextSavePos ++] = keyCode;
		if (k_nextSavePos == 1024) k_nextSavePos = 0;
	} else {
		processCode(keyCode);
	}
}
void k_startBufferingCodes() {
	//k_stopBufferingCodes();
	buffering = 1;
}
void k_stopBufferingCodes() {
	//Procesa todo lo buffereado
	while (k_nextSavePos != k_nextReadPos) {
		processCode(k_buffer[k_nextReadPos ++]);
		if (k_nextReadPos == 1024) k_nextReadPos = 0;
	}

	buffering = 0;
}


unsigned char k_getNextCode() {
	processNextCode();
	return lastCode;
}
unsigned char k_getNextChar() {
	processNextCode();
	return lastChar;
}
unsigned char k_getLastCode() {
	return lastCode;
}
unsigned char k_getLastChar() {
	return lastChar;
}