; *****************************************************************************
; INDICE
; *****************************************************************************
; MACRO
; -----------------------------------------------------------------------------
; PRINT_FROM_STACK (fila, columna, items, stack_addr, label_addr, label_len 
; 	[label_spacing, read_reverse])
; JUMP_ON_XMM_NOT_LOADED dst_ptr
; -----------------------------------------------------------------------------
; PROCEDIMIENTO
; -----------------------------------------------------------------------------
; _fxsave()
; _fxrstor()
; txt_mode()
; -----------------------------------------------------------------------------
; EXCEPCION 
; -----------------------------------------------------------------------------
; _isr0
; _isr1
; _isr2
; _isr3
; _isr4
; _isr5
; _isr6
; _isr7
; _isr8
; _isr9
; _isr10
; _isr11
; _isr12
; _isr13
; _isr14
; _isr16
; _isr17
; _isr18
; _isr19
; -----------------------------------------------------------------------------
; PROCEDIMIENTO
; -----------------------------------------------------------------------------
; print_regs_halt()
; clearScreen()
; switch_task()
; -----------------------------------------------------------------------------
; INTERRUPCION
; -----------------------------------------------------------------------------
; _isr32 (IRQ0 - TIMER)
; _isr33 (IRQ1 - TECLADO)
; _isr39 (IRQ7 - solo interrupcion espuria)
; _isr46 (IRQ14 - dispositivo maestro en el bus IDE primario)
; _isr47 (IRQ15 - solo interrupcion espuria)
;--  _isr66 (Dormir proceso hasta que se presione la tecla T)
; _isr67 (Dormir tarea)
; _isr68 (Esperar)
; _isr69 (Consultas al driver grafico)
; _isr70 (Memoria dinamica)
; _isr71 (Interfaz con el teclado)
; _isr72 (Interfaz con el file system)
;

BITS 32
%include "macros/macrosmodoprotegido.mac"
%include "macros/v16_macros.mac"

extern fin_intr_pic1
extern fin_intr_pic2
extern IDT_DESC
extern loadXMM
extern kill_current_task

;Excepciones
isr0_s: db 'Divide Error'
isr0_l equ $-isr0_s

isr2_s: db 'NMI Interrupt'
isr2_l equ $-isr2_s

isr3_s: db 'Breakpoint'
isr3_l equ $-isr3_s

isr4_s: db 'Overflow'
isr4_l equ $-isr4_s

isr5_s: db 'BOUND Range Exceeded'
isr5_l equ $-isr5_s

isr6_s: db 'Invalid Opcode'
isr6_l equ $-isr6_s

isr7_s: db 'Device Not Available'
isr7_l equ $-isr7_s

isr8_s: db 'Double Fault'
isr8_l equ $-isr8_s

isr9_s: db 'Coprocessor Segment Overrun'
isr9_l equ $-isr9_s

isr10_s: db 'Invalid TSS'
isr10_l equ $-isr10_s

isr11_s: db 'Segment Not Present'
isr11_l equ $-isr11_s

isr12_s: db 'Stack-Segment Fault'
isr12_l equ $-isr12_s

isr13_s: db 'General Protection'
isr13_l equ $-isr13_s

isr14_s: db 'Page Fault'
isr14_l equ $-isr14_s

isr16_s: db 'x87 FPU Floating-Point Error'
isr16_l equ $-isr16_s

isr17_s: db 'Alignment Check'
isr17_l equ $-isr17_s

isr18_s: db 'Machine Check'
isr18_l equ $-isr18_s

isr19_s: db 'SIMD Floating-Point Exception'
isr19_l equ $-isr19_s

global _isr0
global _isr1
global _isr2
global _isr3
global _isr4
global _isr5
global _isr6
global _isr7
global _isr8
global _isr9
global _isr10
global _isr11
global _isr12
global _isr13
global _isr14
global _isr16
global _isr17
global _isr18
global _isr19

global _isr32
global _isr33
global _isr39
global _isr46
global _isr47

global _isr66
global _isr67
global _isr68

global _isr69
global _isr70
global _isr71
global _isr72

global switch_task

global _fxrstor
global _fxsave

; *****************************************************************************
; MACRO PRINT_FROM_STACK
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Lee DWORDS del stack y los imprime en pantalla en su expresion 
; hexadecimal. Se especifica una fila y columna de inicio, una cantidad de
; elementos a leer del stack y la direccion desde donde empieza a leer.
; 	Opcionalmente se pueden una direccion con etiquetas (todas de longitud 
; N y consecutivas) para mostrar junto a cada valor que se lea del stack. Se
; modificar el un espaciado entre etiqueta y valor.
; 	Tambien es posible indicar que los valores del stack se lean al reves 
; (es decir desde STACK + 4 * CANT_VALORES hasta STACK).
;
; NOTAS: 
;	Solo funciona en modo texto. 
;	No usar ECX en ningun campo.
;
; -----------------------------------------------------------------------------
; Argumentos (6 a 8)
; -----------------------------------------------------------------------------
; 	%1 = Fila
; 	%2 = Columna
; 	%3 = Cantidad de cosas a imprimir
; 	%4 = Direccion en el stack para pasar Dwords a texto
; 	%5 = Direccion de las etiquetas para cada elemento del sp
; 	%6 = Longitud de cada etiqueta (si es 0 ignora %5 y %7)
; 	%7 = Separacion entre etiqueta y el valor [1 por defecto]
; 	%8 = Si los datos se imprimen como aparecen en el stack (0) o a la 
; inversa [0 por defecto]
;
; -----------------------------------------------------------------------------
; Ejemplo de uso
; -----------------------------------------------------------------------------
; 	Imprimir el estado de todos los registros (suponiendo que eax_s tiene la 
; direccion de la primera etiqueta):
;
; > pushad
; > PRINT_FROM_STACK 3, 1, 8, esp, eax_s, 3, 5, 1
; > add esp, 32
;
dw_str: dq 0
backup: dd 0
%macro PRINT_FROM_STACK 6-8 1, 0
	mov [backup], ecx
	%assign row %1
	%assign col_str %2
	%if %6 > 0
		%assign col_num %2 + %6 + %7
	%else
		%assign col_num %2
	%endif
	
	%if %8 = 0
		%assign stack_i 0
		%assign stack_inc 1
	%else
		%assign stack_i %3 - 1
		%assign stack_inc -1
	%endif
	%assign label_i 0
	
	%rep %3
		%if %6 > 0
			;Lee la etiqueta
			mov ecx, %5
			lea ecx, [ecx + %6 * label_i]
			
			;Imprime el texto
			IMPRIMIR_TEXTO ecx, %6, 0x0F, row, col_str
		%endif
		
		;Lee el numero
		mov ecx, %4
		lea ecx, [ecx + 4 * stack_i]
		mov ecx, [ecx]
		
		;Imprime el numero
		DWORD_TO_HEX ecx, dw_str
		IMPRIMIR_TEXTO dw_str, 8, 0x0F, row, col_num
		
		%assign label_i label_i + 1
		%assign stack_i stack_i + stack_inc
		%assign row row + 1
	%endrep
	
	mov ecx, [backup]
%endmacro

; *****************************************************************************
; MACRO JUMP_ON_XMM_NOT_LOADED
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Si CR0.TS (bit 3) esta en 0 (indicando que hubo un task switch), salta a 
; la posicion de memoria indicada como argumento. En caso contrario, continua 
; luego de la macro.
;
; -----------------------------------------------------------------------------
; Argumentos ()
; -----------------------------------------------------------------------------
; 	%1 = Direccion de memoria a saltar.
;
; -----------------------------------------------------------------------------
; Ejemplo de uso
; -----------------------------------------------------------------------------
; > JUMP_ON_XMM_NOT_LOADED xmm_off
; > ; los xmm estan cargados!
; > ret
; > xmm_off:
; > ; leo los xmm
; > ret
;
%macro JUMP_ON_XMM_NOT_LOADED 1
	push eax
	mov eax, cr0
	test eax, 8
	jz %1
	xor eax, 8
	mov cr0, eax
	pop eax
%endmacro

; *****************************************************************************
; PROCEDIMIENTO _fxsave
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Guarda el estado de los registros que utilizan las tecnologias FPU,
; MMX y SSE en la posicion de memoria indicada como argumento (debe estar 
; alineada a 16 bits o saltara una excepcion GP). Todos los registros de 
; proposito general se preservan (salvo EAX)
;
; -----------------------------------------------------------------------------
; Argumentos (1)
; -----------------------------------------------------------------------------
; 	fxData *mem = puntero alineado a 16 bits que contiene 512 bytes de memoria
; donde se guardara el contexto actual de los registros FPU/MMX/SSE.
;
; Ejemplo de uso
; -----------------------------------------------------------------------------
; 	Luego de cambiar de tarea y recibir una interrupcion luego de ejecutar una
; instruccion de SSE:
;
; > if (previous_task != actual_task) {
; >		_fxsave(tasks[previous_task].x_mem);
; >		_fxrstor(tasks[actual_task].x_mem);
; > }
;
_fxsave:
	mov eax, [esp + 4]
	fxsave [eax]
	ret


; *****************************************************************************
; PROCEDIMIENTO _fxrstor
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Recupera el estado de los registros que utilizan las tecnologias FPU,
; MMX y SSE con la informacion que esta en la posicion de memoria indicada como 
; argumento (debe estar alineada a 16 bits o saltara una excepcion GP).  Todos 
; los registros de proposito general se preservan (salvo EAX)

; -----------------------------------------------------------------------------
; Argumentos (1)
; -----------------------------------------------------------------------------
; 	fxData *mem = puntero alineado a 16 bits que contiene 512 bytes de memoria
; desde donde se cargara el contexto de los registros FPU/MMX/SSE.
;
; Ejemplo de uso
; -----------------------------------------------------------------------------
; 	Luego de cambiar de tarea y recibir una interrupcion luego de ejecutar una
; instruccion de SSE:
;
; > if (previous_task != actual_task) {
; >		_fxsave(tasks[previous_task].x_mem);
; >		_fxrstor(tasks[actual_task].x_mem);
; > }
;
_fxrstor:
	mov eax, [esp + 4]
	fxrstor [eax]
	ret

; *****************************************************************************
; PROCEDIMIENTO txt_mode
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Pasa a modo texto 80x25 de 16 colores (13h). No modifica ningun registro
; de proposito general, ni los segmentos ds, es, fs y gs.
;
; -----------------------------------------------------------------------------
; Argumentos (-)
; -----------------------------------------------------------------------------
; Ejemplo de uso
; -----------------------------------------------------------------------------
; > mov eax, 0xB8000
; > call txt_mode
; > mov word [eax], 0x0F0F
;
txt_mode:
	pushad 
	push ds
	push es
	push fs
	push gs
	
	V16_INT 0x10, 'ax', 3
	
	pop gs
	pop fs
	pop es
	pop ds
	popad
	ret
	
; *****************************************************************************
; EXCEPCION 0, 1, 2, 3, 4, 5 y 6
; -----------------------------------------------------------------------------;
;
_isr0:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr0_s, isr0_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr1:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr0_s, isr0_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr2:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr2_s, isr2_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr3:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr3_s, isr3_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr4:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr4_s, isr4_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr5:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr5_s, isr5_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr6:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr6_s, isr6_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
; *****************************************************************************
; EXCEPCION 7 (Device not available)
; -----------------------------------------------------------------------------;
; Descripcion
; -----------------------------------------------------------------------------;
; 	Chequea si no esta cargado el contexto de las tecnologias FPU/MMX/SSE. 
; Si este es el caso, hace lo necesario para que se puedan utilizar las 
; instrucciones de estas tecnologias y devuelve el control del procesador a la 
; instruccion que genero la excepcion.
; 	En caso contrario, continua como todas las otras excepciones.
;
_isr7:
	JUMP_ON_XMM_NOT_LOADED ._err
	pushad
	call loadXMM
	popad
	
	iretd
._err:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr7_s, isr7_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
; *****************************************************************************
; EXCEPCION 8, 9, 10, 11, 12, 13, 14, 16, 17, 18 y 19
; -----------------------------------------------------------------------------;
;
_isr8:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr8_s, isr8_l, 0x0F, 1, 0
	jmp print_regs_halt_err
	
_isr9:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr9_s, isr9_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr10:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr10_s, isr10_l, 0x0F, 1, 0
	jmp print_regs_halt_err
	
_isr11:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr11_s, isr11_l, 0x0F, 1, 0
	jmp print_regs_halt_err
	
_isr12:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr12_s, isr12_l, 0x0F, 1, 0
	jmp print_regs_halt_err
	
_isr13:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr13_s, isr13_l, 0x0F, 1, 0
	jmp print_regs_halt_err
	
_isr14:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr14_s, isr14_l, 0x0F, 1, 0
	jmp print_regs_halt_err

_isr16:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr16_s, isr16_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr17:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr17_s, isr17_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr
	
_isr18:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr18_s, isr18_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr

_isr19:
	call kill_current_task
	call txt_mode
	call clearScreen
	IMPRIMIR_TEXTO isr19_s, isr19_l, 0x0F, 1, 0
	jmp print_regs_halt_noerr

; *****************************************************************************
; PROCEDIMIENTO print_regs_halt
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Imprime en pantalla la informacion de los registros de proposito general,
; los selectores de segmento, los registros de control, el registro EFLAGS, 
; los ultimos seis valores del stack y el backtrace.
; 	Este procedimiento espera que sea llamado luego de una excepcion sin
; cambio de pila, pudiendose ejecutarse por una excepcion con codigo de error
; utilizando preg_regs_halt_err, o bien por una que no lo tiene usando 
; preg_regs_halt_noerr.
;	Luego de presentar la informacion en pantalla, procede a detener el 
; procesador mediante la instruccion HLT.
; -----------------------------------------------------------------------------
; Argumentos (-)
;
eip_s: db 'EIP'
eax_s: db 'EAX'
ecx_s: db 'ECX'
edx_s: db 'EDX'
ebx_s: db 'EBX'
esp_s: db 'ESP'
ebp_s: db 'EBP'
esi_s: db 'ESI'
edi_s: db 'EDI'

cr0_s: db 'CR0'
cr2_s: db 'CR2'
cr3_s: db 'CR3'
cr4_s: db 'CR4'

cs_s: db 'CS'
ds_s: db 'DS'
es_s: db 'ES'
fs_s: db 'FS'
gs_s: db 'GS'
ss_s: db 'SS'
eflags_s: db 'EFLAGS'
error_s: db 'ERROR'

stack_s: db 'stack'
stack_l equ $-stack_s

trace_s: db 'trace'
trace_l equ $-trace_s

aux: dq 0

eip_stack: dd 0
error_value: dd 0
print_regs_halt_err:
	push eax
	mov eax, [esp + 4]
	mov [error_value], eax
	pop eax
	
	mov [eip_stack], esp
	add dword [eip_stack], 4
	jmp print_regs_halt
	
print_regs_halt_noerr:
	mov dword [error_value], 0
	mov [eip_stack], esp
	
print_regs_halt:
	;Imprimo los registros de proposito general
	pushad
	PRINT_FROM_STACK 3, 1, 8, esp, eax_s, 3, 5, 1
	add esp, 32
	
	;Imprimo los de control
	mov eax, cr0
	push eax
	mov eax, cr2
	push eax
	mov eax, cr3
	push eax
	mov eax, cr4
	push eax
	PRINT_FROM_STACK 12, 1, 4, esp, cr0_s, 3, 5, 1
	add esp, 16
	
	;EFLAGS
	mov esi, [eip_stack]		;Pudo venir con error code!! - Se que aca tengo el EIP
	IMPRIMIR_TEXTO eflags_s, 6, 0x0F, 17, 1
	mov eax, [esi + 8]
	DWORD_TO_HEX eax, aux
	IMPRIMIR_TEXTO aux, 8, 0x0F, 17, 9
	
	;ERROR CODE
	mov eax, [error_value]
	test eax, eax
	IMPRIMIR_TEXTO error_s, 5, 0x0F, 1, 40
	
	;Imprime la parte del selector
	shr eax, 3
	jz .no_error_code
	DWORD_TO_HEX eax, aux
	IMPRIMIR_TEXTO aux+4, 4, 0x0F, 1, 48
	
	mov byte [aux], '|'
	IMPRIMIR_TEXTO aux, 1, 0x0F, 1, 53
	
	;Imprime los tres primeros bits del error
	mov eax, [error_value]
	and eax, 0x07
	add eax, '0'
	mov [aux], al
	IMPRIMIR_TEXTO aux, 1, 0x0F, 1, 55
.no_error_code:
	
	;Segments!
	xor eax, eax
	mov eax, cs
	push eax
	mov eax, ds
	push eax
	mov eax, es
	push eax
	mov eax, fs
	push eax
	mov eax, gs
	push eax
	mov eax, ss
	push eax
	PRINT_FROM_STACK 3, 25, 6, esp, cs_s, 2, 6, 1
	add esp, 24
	
	;stack
	IMPRIMIR_TEXTO stack_s, stack_l, 0x0F, 10, 25
	PRINT_FROM_STACK 12, 25, 6, esp, 0, 0, 0, 0
	
	;backtrace
	IMPRIMIR_TEXTO trace_s, trace_l, 0x0F, 10, 44
	
	mov eax, [eip_stack]
	push dword [eax]
	
	mov ecx, 5
	mov eax, ebp
	test eax, eax
	jz backtrace_end
	
	.create_backtrace:
		push dword [eax + 4]
		mov eax, [eax]
		test eax, eax
	loopnz .create_backtrace
	test ecx, ecx
	jz backtrace_ready
	xor eax, eax
	.backtrace_fill:
		push eax
	loop .backtrace_fill
backtrace_ready:
	PRINT_FROM_STACK 12, 44, 6, esp, 0, 0, 0, 1
	add esp, 24
	
backtrace_end:
	cmp byte [timing_stat], 17
	jae error_animation
	hlt

%define ANIM_MS 1000
cycles: dd 0
ms: dd 0
error_animation:
	;Leo los ciclos actuales
	rdtsc
	mov [act_ts], eax
	mov [act_ts + 4], edx
	movq xmm0, [act_ts]
	
	;Tenemos la cantidad de ciclos por ms
	movq xmm7, [tmr_cycle]
	
	_big_wait_loop:
		mov ecx, 100
		.wait_loop:
		loop .wait_loop
		
		;Leo el TS actual
		rdtsc
		mov [act_ts], eax
		mov [act_ts + 4], edx
		movq xmm1, [act_ts]
		
		;Saco la diferencia y sumo con la cantidad de ciclos por ms
		psubq xmm1, xmm0
		psubq xmm1, xmm7
		
		psrldq xmm1, 4
		movd eax, xmm1
		cmp eax, 0
	jl _big_wait_loop ;Si eax < 0 => (act_ts - prev_ps) - tmr_cycle < 0 => eax < tmr_cycle
	
	;Reemplazo el valor viejo
	rdtsc
	mov [act_ts], eax
	mov [act_ts + 4], edx
	movq xmm0, [act_ts]
	
	inc dword [ms]
	cmp dword [ms], ANIM_MS		;si ms < ANIM_MS, repito lo de arriba
	jb _big_wait_loop
	
	mov dword [ms], 0

anim_cycle:
	inc dword [cycles]
	mov eax, [cycles]
	DWORD_TO_HEX eax, aux
	IMPRIMIR_TEXTO aux, 8, 0x0F, 0, 0
	jmp _big_wait_loop
	

%define SCR_WIDTH 80
anim_columns:	;x y size spd
	%rep SCR_WIDTH
		db 0, 0, 0, 0
	%endrep
	
	
	
	
	
	
	
; *****************************************************************************
; PROCEDIMIENTO clearScreen
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Considera que esta en modo texto de 80x25 con 16 colores. Ignora la 
; primera y ultima linea, deja en color negro al resto. No modifica ningun
; registro de proposito general.
;
; -----------------------------------------------------------------------------
; Argumentos (-)
; -----------------------------------------------------------------------------
; Ejemplo de uso
; -----------------------------------------------------------------------------
; > mov eax, 0xB8000
; > call clearScreen
; > mov word [eax], 0x0F0F
;
clearScreen:
	pushad
	xor eax, eax
	mov ecx, 80*23
	mov edi, 0xB80A0
	rep stosw
	
	popad
	ret
	
; *****************************************************************************
; PROCEDIMIENTO switch_task
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Pide al scheduler la siguiente tarea. Si no hay ninguna o es la misma, 
; no hace nada. En caso contrario, conmuta a ella.
;
; -----------------------------------------------------------------------------
; Argumentos (-)
;
extern proximo_indice
extern switch_task_return
	
switch_task:
	;xchg bx, bx
	call proximo_indice
	cmp ax, 0
	jz switch_task_ret
	
	str bx
	cmp bx, ax
	jz switch_task_ret

	mov [jmp_task + 4], ax
	jmp far [jmp_task]
	
switch_task_ret:
	call switch_task_return
	ret

jmp_task: dd 0
	dw 0


; *****************************************************************************
; INTERRUPCION 32 (IRQ0 - TIMER)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Las primeras veces que es llamada, calcula un promedio de ciclos de 
; procesador por interrupcion de TIMER.
;	Procede a informar al scheduler que ocurrio un tick. Si es necesario, 
; conmuta a la siguiente tarea.
;
extern doStep

prev_ts: dq 0
act_ts: dq 0
tmr_cycle: dq 0
timing_stat: db 0

xmm0_backup: dq 0, 0
xmm1_backup: dq 0, 0

_isr_count: dd 0
_isr32:
	pushad
	
	cmp byte [timing_stat], 17
	jae _isr32_cont
	
	inc byte [timing_stat]
	cmp byte [timing_stat], 1
	jne _isr32_chk_end_tmr			;Si es la primera vez que entra, guarda el TS
	
	;Leo el timestamp actual
	rdtsc
	mov [prev_ts], eax
	mov [prev_ts + 4], edx
	jmp _isr32_cont
	
_isr32_chk_end_tmr:
	cmp byte [timing_stat], 17	;Si es la ultima vez que va a entrar, calcula
	jne _isr32_cont
	
	rdtsc
	mov [act_ts], eax
	mov [act_ts + 4], edx
	
	;Calcula la diferencia entre interrupciones de timer
	;y lo guarda en tmr_cycle
	
	;Uso SSE, estallara todo?
	movdqu [xmm0_backup], xmm0
	movdqu [xmm1_backup], xmm1
	movq xmm0, [prev_ts]
	movq xmm1, [act_ts]
	psubq xmm1, xmm0
	psrlq xmm1, 4				;Divide por 16
	movq [tmr_cycle], xmm1
	movdqu xmm0, [xmm0_backup]
	movdqu xmm1, [xmm1_backup]
	
	;mov eax, 0
	;mov cs, ax

_isr32_cont:
	;Aviso que se atendio la interrupcion
	call fin_intr_pic1
	
	;Actualizo las tareas y hace switch si hace falta
	call doStep
	call switch_task

_isr32_ret:
	popad

	iretd

; *****************************************************************************
; INTERRUPCION 33 (IRQ1 - TECLADO)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Obtiene la tecla que fue presionada y despierta a las tareas que se fueron
; a dormir hasta la presion de dicha tecla.
;
;extern warnKeyPress
extern k_gotInput
_isr33:
	pushad
	call fin_intr_pic1
	xor eax, eax
	in al, 0x60
	
	push eax
	;call warnKeyPress
	call k_gotInput
	add esp, 4
	
	popad
	iretd
	
; *****************************************************************************
; INTERRUPCION 39 (IRQ7) y 47 (IRQ15)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	El kernel no contempla estas interrupciones al considerar que no son
; necesarias. Sin embargo, puede ocurrir que el PIC envie interrupciones 
; espurias a estas dos IRQs, por lo que deben tener una entrada en la IDT o
; sufrir una excepcion GP.
;
_isr39:
	iretd
	
_isr47:
	iretd

; *****************************************************************************
; INTERRUPCION 46 (IRQ14 - Primary IDE master device)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Se recibe luego de una operacion del dispositivo. Como se usa pooling no
; es necesario utilizarla.
;	
_isr46:
	call fin_intr_pic2
	iretd

; *****************************************************************************
; -- INTERRUPCION 66 (Dormir proceso hasta que se presione la tecla T)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Pone a dormir la tarea actual hasta que la tecla que lo despierte sea la T.
;
;extern sleep
;extern wakingKeyCode
;extern resetWakingKeyCode
;
;_isr66:
;	pushad
;_isr66_loop:
;	call sleep
;	
;	call wakingKeyCode
;	cmp ax, 20
;	jnz _isr66_loop
;_isr66_ret:
;	call resetWakingKeyCode
;	
;	popad
;	iretd
	
; *****************************************************************************
; INTERRUPCION 67 (Dormir tarea)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Pone a dormir la tarea actual la cantidad de ticks que indique EAX. 
; Conmuta al siguiente proceso.
;
extern sleepTask
extern waitTask
extern setTaskActiveTime

_isr67:
	pushad
	
	push eax
	call sleepTask
	add esp, 4
	
	popad
	iretd

; *****************************************************************************
; INTERRUPCION 68 (Esperar)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Deja la tarea en un loop hasta que pasen los ticks indicados en EAX. No
; conmuta al siguiente proceso.
;
_isr68:
	pushad
	
	push eax
	sti			;Pueden interrumpir
	call waitTask
	cli
	add esp, 4
	
	
	
	popad
	iretd
	
; *****************************************************************************
; INTERRUPCION 69 (Consultas al driver grafico)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	Realiza consultas al driver grafico segun el valor en EAX.
;
; 	EAX = 1h
;		Devuelve en EAX el ancho en pixels de la pantalla
; 	EAX = 2h
;		Devuelve en EAX el alto en pixels de la pantalla
; 	EAX = 3h
;		Devuelve en EAX la posicion de la memoria de video donde escribir
;	EAX = 4h
;		Dibuja usando un solo color lo que indique la pila apuntada por EBX:
;			EBX + 00 = (unsigned char *) Puntero a la informacion
;			EBX + 04 = (unsigned int) Cantidad de bytes a leer
; 			EBX + 08 = (unsigned int) Cuantos bits por pixel (4 u 8)
; 			EBX + 12 = (unsigned int) Posicion en X
; 			EBX + 16 = (unsigned int) Posicion en Y
; 			EBX + 20 = (unsigned char) Shiftear al valor esta cantidad
; 			EBX + 24 = (unsigned char) Sumar al valor esta cantidad
; 			EBX + 28 = (unsigned char) Sumar al color esta cantidad si 
;				el byte de color luego de shiftear es distinto de cero
;	EAX = 6h
;		Imprime un string terminado con 0 segun lo que indica la pila 
;			apuntada por EBX:
;			EBX + 00 = (unsigned char *) Puntero a la informacion
; 			EBX + 04 = (unsigned int) Posicion en X
; 			EBX + 08 = (unsigned int) Posicion en Y
;			EBX + 12 = (unsigned char) Color
;			EBX + 16 = (unsigned char) Un solo color (0) o RGB (otro valor)
;		Devuelve en EAX la cantidad de pixels corridos en X donde se deberia
;			seguir imprimiendo
;
extern getScreenWidth
extern getScreenHeight
extern getVideoMemAddr

_isr69_eax: dd 0
_isr69:
	pushad
	mov [_isr69_eax], eax
	
	;Segun el valor de EAX, es lo que se pide
	cmp eax, 0x01		;get screen width
	je .get_screen_width
	cmp eax, 0x02		;get screen height
	je .get_screen_height
	cmp eax, 0x03		;video memory address
	je .get_video_mem_addr
	cmp eax, 0x04		;draw byte
	je .draw_byte
	cmp eax, 0x06		;draw string
	je .draw_string
	jmp .ret
	
.get_screen_width:
	call getScreenWidth
	mov [_isr69_eax], eax
	jmp .ret
	
.get_screen_height:
	call getScreenHeight
	mov [_isr69_eax], eax
	jmp .ret

.get_video_mem_addr:
	call getVideoMemAddr
	mov [_isr69_eax], eax
	jmp .ret
	
extern drawByte

%define data_ptr ebx
%define data_len ebx + 4
%define data_size ebx + 8
%define data_x ebx + 12
%define data_y ebx + 16
%define data_shift ebx + 20
%define data_sum ebx + 24
%define data_sum_cond ebx + 28
.draw_byte:
	;cmp dword [data_size], 4
	;je .draw_monochrome_4
	;jmp .draw_monochrome_8
	
	.draw_monochrome_4:
		
		;backupeo los xmm que voy a usar
		PUSH_XMM xmm0, xmm1, xmm2, xmm3, xmm4, xmm5
		sub esp, 0x10
		mov eax, esp
		
		mov esi, [data_ptr]
		mov edi, [data_len]
		shr edi, 3
		
		;Preparo la pila
		push dword [data_y]
		push dword [data_x]
		push dword 16
		push eax
		
		;Suma condicional
			mov cl, [data_sum_cond]
			cmp cl, 0
			jz .no_sum_cond
			mov al, cl
			mov ah, cl
			bswap eax
			mov al, cl
			mov ah, cl
			
			movd xmm5, eax
			pslldq xmm5, 3
			psrldq xmm5, 3
			pshufd xmm5, xmm5, 0
		.no_sum_cond:
			
		;Suma al color
			pxor xmm4, xmm4
			mov cl, [data_sum]
			cmp cl, 0
			jz .no_sum
			mov al, cl
			mov ah, cl
			bswap eax
			mov al, cl
			mov ah, cl
			
			movd xmm4, eax
			pslldq xmm4, 3
			psrldq xmm4, 3
			pshufd xmm4, xmm4, 0
		.no_sum:
		
		sti

		pxor xmm2, xmm2	; <- no se usa xmm2
		test edi, edi
		jz .draw_byte_less_8
		.draw_byte_loop:
			pxor xmm1, xmm1
			movq xmm0, [esi]			;xmm0_lq = n15 n14 | .. | n1 n0
			punpcklbw xmm0, xmm1		;xmm0 = 0 | n15 n14 | 0 | n13 n12 | ... | n1 n0
			
			mov eax, 0xF0F0F0F0
			movd xmm1, eax
			pshufd xmm1, xmm1, 0
			pand xmm1, xmm0
			pxor xmm0, xmm1
			pslldq xmm1, 1
			psrlw xmm1, 4
			pxor xmm0, xmm1
			
			psllw xmm0, 7
			
			;Suma al color condicional
			pxor xmm3, xmm3
			pcmpeqb xmm3, xmm0
			pandn xmm3, xmm5
			paddusb xmm0, xmm3
			
			;Suma al color
			paddusb xmm0, xmm4
			
			mov eax, esp
			add eax, 16
			movdqu [eax], xmm0
			call drawByte
			
			add esi, 8
			add dword [esp + 8], 16
			
			dec edi
			jnz .draw_byte_loop
			
		.draw_byte_less_8:
			;Ahora dibujo de a un byte (no testeado)
			
			mov edi, [data_len]
			and edi, 0x7
			
			test edi, edi
			jz .draw_byte_4_ret
			
			mov edx, esp
			add edx, 16
			
			xor eax, eax
			mov [esp + 4], edi
			.draw_byte_loop_byte:
				mov al, [esi + eax]
				mov [edx + eax], al
				
				inc eax
				dec edi
				jnz .draw_byte_loop_byte
				
			call drawByte
		
	.draw_byte_4_ret:
		add esp, 16
		add esp, 0x10
		
		cli
		POP_XMM xmm0, xmm1, xmm2, xmm3, xmm4, xmm5
		jmp .ret
		
	.draw_byte_8:
		jmp .ret

extern drawStringByte
extern drawStringRGB
%define data_ptr ebx
%define data_x ebx + 4
%define data_y ebx + 8
%define data_color ebx + 12
%define data_single ebx + 16
.draw_string:
		push dword [data_y]
		push dword [data_x]
		push dword [data_color]
		push dword [data_ptr]
		cmp byte [data_single], 0x00
		jne .draw_string_rgb
			sti
			call drawStringByte
			cli
			mov [_isr69_eax], eax
			
			add esp, 16
			jmp .ret
		.draw_string_rgb:
			sti
			call drawStringRGB
			cli
			mov [_isr69_eax], eax
			
			add esp, 16
			jmp .ret
.ret:
	popad
	mov eax, [_isr69_eax]
	iretd


; *****************************************************************************
; INTERRUPCION 70 (Peticion de memoria dinamica)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	EBX = 0
;		Obtiene tantas paginas de memoria como haga falta para ocupar lo pedido 
;		en EAX. Devuelve en EAX la posicion de memoria que apunta al primer byte o
; 		NULL (0) si no se pudo obtener la memoria pedida.
;	EBX = 1
;		Libera la memoria dinamica asignada a la direccion pasada en EAX.
;
extern mmu_malloc
extern mmu_free
extern act_pid

_isr70_eax: dd 0
_isr70:
	pushad
	
	cmp ebx, 0
	je .malloc
	cmp ebx, 1
	je .free
	jmp .ret
	
.malloc
	push eax
	push dword [act_pid]
	call mmu_malloc
	add esp, 8
	mov [_isr70_eax], eax
	jmp .ret
	
.free
	push eax
	push dword [act_pid]
	call mmu_free
	add esp, 8
	mov dword [_isr70_eax], 0
	jmp .ret
	
.ret:
	popad
	
	mov eax, [_isr70_eax]
	iretd


; *****************************************************************************
; INTERRUPCION 71 (Interfaz con el teclado)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	EBX = 0
;		Pide el foco del teclado.
;	EBX = 1
;		Pide el siguiente codigo buffereado
;	EBX = 2
;		Pide el siguiente caracter buffereado
;	EBX = 3
;		Pide el ultimo codigo procesado
;	EBX = 4
;		Pide el ultimo caracter procesado
;
extern s_getKeyboardFocus
extern s_getKeyboardNextCode
extern s_getKeyboardNextChar
extern s_getKeyboardLastCode
extern s_getKeyboardLastChar

_isr71_eax: dd 0
_isr71:
	pushad
	
	cmp ebx, 0
	je .focus
	cmp ebx, 1
	je .nextcode
	cmp ebx, 2
	je .nextchar
	cmp ebx, 3
	je .lastcode
	cmp ebx, 4
	je .lastchar
	jmp .ret
	
.focus
	call s_getKeyboardFocus
	mov [_isr71_eax], eax
	jmp .ret
	
.nextcode
	call s_getKeyboardNextCode
	mov [_isr71_eax], eax
	jmp .ret

.nextchar
	call s_getKeyboardNextChar
	mov [_isr71_eax], eax
	jmp .ret

.lastcode
	call s_getKeyboardLastCode
	mov [_isr71_eax], eax
	jmp .ret

.lastchar
	call s_getKeyboardLastChar
	mov [_isr71_eax], eax
	jmp .ret
	
.ret:
	popad
	
	mov eax, [_isr71_eax]
	iretd



; *****************************************************************************
; INTERRUPCION 72 (Interfaz con el file system)
; -----------------------------------------------------------------------------
; Descripcion
; -----------------------------------------------------------------------------
; 	EAX = 0
;		Pide informacion sobre un directorio, conteniendo en EBX el path al 
;		mismo, y en ECX la direccion a la posicion de memoria de un
;		directory_contents donde depositarla.
;	EAX = 1
;		Pide cargar el archivo cuyo path esta en EBX como un proceso y 
;		ponerlo a correr.
;	EAX = 2
;		Pide cargar los archivos cuyo path estan en EBX y ECX como
;		procesos y ponerlos a correr.
;
extern get_directory_contents
extern run_files_as_program
extern run_file_as_program

_isr72_eax: dd 0
_isr72:
	pushad
	
	cmp eax, 0
	je .info
	cmp eax, 1
	je .ld
	cmp eax, 2
	je .ld2
	jmp .ret
	
.info
	push ecx
	push ebx
	call get_directory_contents
	add esp, 8
	mov [_isr72_eax], eax
	jmp .ret
	
.ld
	push ebx
	call run_file_as_program
	add esp, 4
	mov [_isr72_eax], eax
	jmp .ret

.ld2
	push ecx
	push ebx
	call run_files_as_program
	add esp, 8
	mov [_isr72_eax], eax
	jmp .ret

	
.ret:
	popad
	
	mov eax, [_isr72_eax]
	iretd
