#include "mmu.h"
#include "i386.h"
#include "../common/utils.h"
#include "../drivers/graphic.h"

unsigned int userPage;
unsigned int kernelPage;

uint debugged = 0;

task_mmu_list mmu_info;
mmu_recycle_list *recycled_user_pages;
mmu_recycle_list *recycled_kernel_pages;

mmu_recycle_list *free_mmu_recycle_list;
task_alloc_desc *free_task_alloc_desc;
task_phys_stat *free_task_phys_stat;
task_mmu_list *free_task_mmu_list;
task_alloc *free_task_alloc;

void inicializar_mmu() {
	scr_printf("%C[KERNEL]%C Inicializando MMU\n", YELLOW, WHITE);
	kernelPage = INICIO_PAGINAS_KERNEL - TAMANO_PAGINA;
	userPage = INICIO_PAGINAS_USUARIO - TAMANO_PAGINA;
	
	unsigned int dir = pagina_libre_kernel();
	unsigned int *dirPtr = (unsigned int *) dir;
	unsigned int tab = pagina_libre_kernel();
	unsigned int *tabPtr = (unsigned int *) tab;
	dirPtr[0] = tab + 3;
	
	int i = 0;
	while (i < 1024) {
		tabPtr[i] = 0;
		
		if (i < 512) tabPtr[i] = (i << 12) + 3;
		if (i > 0)	dirPtr[i] = 0;
		i ++;
	}
	
	lcr3(dir);
	
	//Inicializo la lista enlazada de tareas
	mmu_info.pid = IDLE_PID;
	mmu_info.space_used = 0;
	mmu_info.alloc = NULL;
	mmu_info.pages = NULL;
	mmu_info.next = NULL;

	//Asi como las paginas fisicas que se dejaron de usar y los recursos
	recycled_user_pages = NULL;
	recycled_kernel_pages = NULL;
	free_mmu_recycle_list = NULL;
	free_task_alloc_desc = NULL;
	free_task_phys_stat = NULL;
	free_task_mmu_list = NULL;
	free_task_alloc = NULL;
	
	//Y busco memoria para todos los recursos de MMU libres
	free_mmu_recycle_list = (mmu_recycle_list *)
		get_more_resources(sizeof(mmu_recycle_list));
	free_task_alloc_desc = (task_alloc_desc *)
		get_more_resources(sizeof(task_alloc_desc));
	free_task_phys_stat = (task_phys_stat *)
		get_more_resources(sizeof(task_phys_stat));
	free_task_mmu_list = (task_mmu_list *)
		get_more_resources(sizeof(task_mmu_list));
	free_task_alloc = (task_alloc *)
		get_more_resources(sizeof(task_alloc));

	//Mapea la memoria de video
	mapear_memoria_video(dir);
}
void *get_more_resources(int size) {
	uint mem = pagina_libre_kernel();
	uint amount = 0x1000 / size;
	uint i;
	
	int *data = NULL;
	
	for (i = 0; i < amount; i ++, mem += size) {
		int* memdata = (int *) mem;
		memdata[0] = (int) data;	//struct->next = (int) data
		data = memdata;
	}
	
	return (void*) data;
}

#define get_free_resource(TYPE) \
	get_free_##TYPE()
	
#define release_resource(TYPE, resource) \
	release_##TYPE(resource)

mmu_recycle_list* get_free_mmu_recycle_list () {
	if (free_mmu_recycle_list == NULL)
		free_mmu_recycle_list = (mmu_recycle_list *)
			get_more_resources(sizeof(mmu_recycle_list));
	
	mmu_recycle_list *ret = free_mmu_recycle_list;
	free_mmu_recycle_list = free_mmu_recycle_list->next;
	return ret;
}
task_alloc_desc* get_free_task_alloc_desc () {
	if (free_task_alloc_desc == NULL)
		free_task_alloc_desc = (task_alloc_desc *)
			get_more_resources(sizeof(task_alloc_desc));

	task_alloc_desc *ret = free_task_alloc_desc;
	free_task_alloc_desc = free_task_alloc_desc->next;
	return ret;
}
task_phys_stat* get_free_task_phys_stat () {
	if (free_task_phys_stat == NULL)
		free_task_phys_stat = (task_phys_stat *)
			get_more_resources(sizeof(task_phys_stat));

	task_phys_stat *ret = free_task_phys_stat;
	free_task_phys_stat = free_task_phys_stat->next;
	return ret;
}
task_mmu_list* get_free_task_mmu_list () {
	if (free_task_mmu_list == NULL)
		free_task_mmu_list = (task_mmu_list *)
			get_more_resources(sizeof(task_mmu_list));

	task_mmu_list *ret = free_task_mmu_list;
	free_task_mmu_list = free_task_mmu_list->next;
	return ret;
}
task_alloc* get_free_task_alloc () {
	if (free_task_alloc == NULL)
		free_task_alloc = (task_alloc *)
			get_more_resources(sizeof(task_alloc));

	task_alloc *ret = free_task_alloc;
	free_task_alloc = free_task_alloc->next;
	return ret;
}


void release_mmu_recycle_list(mmu_recycle_list *resource) {
	resource->next = free_mmu_recycle_list;
	free_mmu_recycle_list = resource;
}
void release_task_alloc_desc(task_alloc_desc *resource) {
	resource->next = free_task_alloc_desc;
	free_task_alloc_desc = resource;
}
void release_task_phys_stat(task_phys_stat *resource) {
	resource->next = free_task_phys_stat;
	free_task_phys_stat = resource;
}
void release_task_mmu_list(task_mmu_list *resource) {
	resource->next = free_task_mmu_list;
	free_task_mmu_list = resource;
}
void release_task_alloc(task_alloc *resource) {
	resource->next = free_task_alloc;
	free_task_alloc = resource;
}


unsigned int pagina_libre_kernel() {
	if (recycled_kernel_pages != NULL) {
		mmu_recycle_list *recycled = recycled_kernel_pages;
		
		uint ret = recycled_kernel_pages->page;
		recycled_kernel_pages = recycled_kernel_pages->next;
		
		release_resource(mmu_recycle_list, recycled);
		
		return ret;
	}
	return kernelPage += TAMANO_PAGINA;
}
unsigned int pagina_libre_usuario() {
	if (recycled_user_pages != NULL) {
		mmu_recycle_list *recycled = recycled_user_pages;
		
		uint ret = recycled_user_pages->page;
		recycled_user_pages = recycled_user_pages->next;
		
		release_resource(mmu_recycle_list, recycled);
		
		return ret;
	}
	return userPage += TAMANO_PAGINA;
}
void recycleUserPage(unsigned int page) {
	mmu_recycle_list *recycled = get_free_resource(mmu_recycle_list);
	recycled->next = recycled_user_pages;
	recycled->page = page;
	recycled_user_pages = recycled;
}
void recycleKernelPage(unsigned int page) {
	mmu_recycle_list *recycled = get_free_resource(mmu_recycle_list);
	recycled->next = recycled_kernel_pages;
	recycled->page = page;
	recycled_kernel_pages = recycled;
}

unsigned int inicializar_dir_usuario() {
	unsigned int dir = pagina_libre_kernel();
	unsigned int *dirPtr = (unsigned int *) dir;
	unsigned int tab = pagina_libre_kernel();
	unsigned int *tabPtr = (unsigned int *) tab;
	dirPtr[0] = tab + 3;
	
	int i = 1;
	tabPtr[0] = 0;	//No quiero mapear la primera pagina (0x0000 - 0x0FFF)
	while (i < 1024) {
		tabPtr[i] = 0;
		
		if (i < 512) tabPtr[i] = (i << 12) + 7;
		if (i < 256) tabPtr[i] = (i << 12) + 3;
		if (i > 0)	dirPtr[i] = 0;
		i ++;
	}

	//Mapea la memoria de video
	mapear_memoria_video(dir);
	
	return dir;
}
unsigned int direccion_fisica(unsigned int mvirt, unsigned int cr3) {
	unsigned int *dirPtr = (unsigned int *) (cr3);
	unsigned int dirOff = (mvirt >> 22) & 0x3FF;
	unsigned int tabOff = (mvirt >> 12) & 0x3FF;
	
	unsigned int dirData = dirPtr[dirOff];
	unsigned int *tabPtr = (unsigned int *) (dirData &~0x3FF);
	
	if (!(dirData & 1)) {
		unsigned int tab = pagina_libre_kernel();
		tabPtr = (unsigned int *) tab;
		
		dirPtr[dirOff] = tab + 3;
		int i = 0;
		while (i < 1024) {
			tabPtr[i] = 0;
			i ++;
		}
	}
	
	if (dirData & 128) return dirData;
	return tabPtr[tabOff] &~ 0x3FF;
}

uint mapear_memoria_video(unsigned int cr3) {
	unsigned int mem = getVideoMemAddr();
	uint ret = obtener_entrada_directorio(mem, cr3);
	mapear_4mb(mem, cr3, mem);

	/*
	unsigned int pages = (getVideoMemorySize() + 0xFFF) >> 12;
	unsigned int i;
	
	if (pages > 0x1000) {
		pages = (pages + 0x3FF) >> 10;
		for (i = 0; i < pages; i ++, mem += 0x4000000) {
			mapear_4mb(mem, cr3, mem);
		}
	} else {
		for (i = 0; i < pages; i ++, mem += 0x1000) {
			mapear_pagina(mem, cr3, mem);
		}
	}
	*/
	return ret;
}
void unmapear_memoria_video(unsigned int cr3, unsigned int dir_data) {
	unsigned int mem = getVideoMemAddr();
	setear_entrada_directorio(mem, cr3, dir_data);
}
void setear_entrada_directorio(unsigned int mvirt, unsigned int cr3, unsigned int data) {
	unsigned int *dirPtr = (unsigned int *) (cr3);
	unsigned int dirOff = (mvirt >> 22) & 0x3FF;
	
	dirPtr[dirOff] = data;
	if (cr3 == rcr3()) tlbflush();
}
uint obtener_entrada_directorio(unsigned int mvirt, unsigned int cr3) {
	unsigned int *dirPtr = (unsigned int *) (cr3);
	unsigned int dirOff = (mvirt >> 22) & 0x3FF;
	
	return dirPtr[dirOff];
}
void mapear_4mb(unsigned int mvirt, unsigned int cr3, unsigned int mfis) {
	unsigned int *dirPtr = (unsigned int *) (cr3);
	unsigned int dirOff = (mvirt >> 22) & 0x3FF;
	
	dirPtr[dirOff] = (mfis &~0x3FFFFF) + (1 | 2 | 128);
	if (cr3 == rcr3()) tlbflush();
}
void mapear_pagina(unsigned int mvirt, unsigned int cr3, unsigned int mfis) {
	unsigned int *dirPtr = (unsigned int *) (cr3);
	unsigned int dirOff = (mvirt >> 22) & 0x3FF;
	unsigned int tabOff = (mvirt >> 12) & 0x3FF;
	
	unsigned int dirData = dirPtr[dirOff];
	unsigned int *tabPtr = (unsigned int *) (dirData &~0x3FF);
	
	if (!(dirData & 1) || (dirData & 128)) {
		unsigned int tab = pagina_libre_kernel();
		tabPtr = (unsigned int *) tab;
		
		dirPtr[dirOff] = tab + 3;
		int i = 0;
		while (i < 1024) {
			tabPtr[i] = 0;
			i ++;
		}
	}
	
	tabPtr[tabOff] = (mfis &~0x3FF) + 3;
	if (cr3 == rcr3()) tlbflush();
}
void unmapear_4mb(unsigned int mvirt, unsigned int cr3) {
	unsigned int *dirPtr = (unsigned int *) (cr3);
	unsigned int dirOff = (mvirt >> 22) & 0x3FF;
	
	dirPtr[dirOff] = 0;
	if (cr3 == rcr3()) tlbflush();
}
void unmapear_pagina(unsigned int mvirt, unsigned int cr3) {
	unsigned int *dirPtr = (unsigned int *) (cr3);
	unsigned int dirOff = (mvirt >> 22) & 0x3FF;
	unsigned int tabOff = (mvirt >> 12) & 0x3FF;
	
	unsigned int dirData = dirPtr[dirOff];
	unsigned int *tabPtr = (unsigned int *) (dirData &~0x3FF);
	
	if (!(dirData & 1) || (dirData & 128)) {
		unsigned int tab = pagina_libre_kernel();
		tabPtr = (unsigned int *) tab;
		
		dirPtr[dirOff] = tab + 3;
		int i = 0;
		while (i < 1024) {
			tabPtr[i] = 0;
			i ++;
		}
	}
	
	tabPtr[tabOff] = 0;
	if (cr3 == rcr3()) tlbflush();
}

void recycleDirectoryMemory(unsigned int dir) {
	unsigned int *dirPtr = (unsigned int *) dir;
	
	int i = 0;
	for (; i < 1024; i++) {
		if ((dirPtr[i] & 1) && ((dirPtr[i] & 128) == 0)) {
			//Si la entrada esta activa y no es de 4mb
			unsigned int tabPage = dirPtr[i] &~0x3FF;
			recycleKernelPage(tabPage);
		}
	}
	recycleKernelPage(dir);
}

task_mmu_list *getTaskMMUDescriptor(uint pid) {
	task_mmu_list *plst = &mmu_info;
	task_mmu_list *nlst = plst->next;
	
	//Chequea si el proceso ya tiene un descriptor asignado
	if (plst->pid == pid)	return plst;
	while (plst->next != NULL) {
		if (nlst->pid == pid) return nlst;
		
		plst = nlst;
		nlst = nlst->next;
	}
	
	//No lo tiene, tengo que crear uno nuevo
	nlst = get_free_resource(task_mmu_list);
	nlst->pid = pid;
	nlst->space_used = 0;
	nlst->alloc = NULL;
	nlst->pages = NULL;
	nlst->next = NULL;
	
	plst->next = nlst;
	return nlst;
}

void *allocate_after(task_mmu_list *tsk_desc, task_alloc *prev_alloc, 
	uint virtual_from, uint size) {
	
	//Hay que cambiar las estructuras de alloc
	task_alloc *alloc = get_free_resource(task_alloc);
	if (prev_alloc == NULL) {
		//Nada antes
		alloc->next = tsk_desc->alloc;
		tsk_desc->alloc = alloc;
	} else {
		//prev_alloc esta antes
		alloc->next = prev_alloc->next;
		prev_alloc->next = alloc;
	}
	
	//Por ahora no uso space para esta estructura
	alloc->virtual_from = virtual_from;
	alloc->size = size;
	//alloc->space = NULL;
	
	tsk_desc->space_used += size;
	
	//Hay que tocar las estructuras de memoria fisica tambien
	uint page = (virtual_from & ~0xFFF);
	uint last_page = (virtual_from + size);
	
	//Mapeo todas las paginas de memoria fisica que haga falta mapear
	task_phys_stat *phys = tsk_desc->pages;
	
	uint addr = virtual_from;
	for (; page < last_page; page += 0x1000) {
		//Busco phys tal que phys->next->virtual_address > page
		//Me quedo con phys, donde virtual_address <= page
		if (phys != NULL) {
			while ((phys->next != NULL) && (phys->next->virtual_address <= page)) {
				phys = phys->next;
			}
		}
	
		if (phys == NULL || phys->virtual_address != page) {
			//La pagina no esta tomada, creo la estructura y la tomo
			task_phys_stat *nphys = get_free_resource(task_phys_stat);
			nphys->page = pagina_libre_usuario();
			nphys->virtual_address = page;
			nphys->space = NULL;
			
			mapear_pagina(page, rcr3(), nphys->page);
			
			//Si no tenia paginas antes, queda esta entrada como primera
			if (phys == NULL) {
				nphys->next = NULL;
				tsk_desc->pages = nphys;
			} else {
				//Pudo pasar que tsk_desc->pages->virual_address > page!
				if (phys->virtual_address > page) {
					//phys viene despues de nphys
					tsk_desc->pages = nphys;
					nphys->next = phys;
				} else {
					//phys->virt_addr < page, entonces nphys viene despues
					nphys->next = phys->next;
					phys->next = nphys;
				}
			}
			
			//Dejo en phys la estructura que tiene virtual_address == page
			phys = nphys;
		}
		
		//Tengo phys->virtual_address == page
		//Ahora lleno phys->space
		uint phys_start = (addr & 0xFFF) + phys->page;
		
		//Busco space tal que space->next->mem_from > phys_start
		//Me quedo con space, donde mem_from <= phys_start
		task_alloc_desc *space = phys->space;
		if (space != NULL) {
			while ((space->next != NULL) && (space->next->mem_from < phys_start)) {
				space = space->next;
			}
			
			task_alloc_desc *nspace = get_free_resource(task_alloc_desc);
			
			//Pudo pasar que phys->space->mem_from > phys_start!
			if (space->mem_from > phys_start) {
				//Entonces, nspace esta antes que space
				phys->space = nspace;
				nspace->next = space;
			} else {
				nspace->next = space->next;
				space->next = nspace;
			}
			
			space = nspace;
		} else {
			//No tenia la estructura antes
			space = get_free_resource(task_alloc_desc);
			phys->space = space;
			
			space->next = NULL;
		}
		
		//Tengo space para setear la informacion
		space->mem_from = phys_start;
		
		
		//Calculo el espacio utilizado en esta pagina de memoria fisica
		//No se va a pisar con nspace->next porque supuestamente
		//virtual_from esta tomado para que se puedan utilizar size
		//bytes consecutivos
		uint end = size + (addr & 0xFFF);
		if (end >= 0x1000) {
			space->space = 0x1000 - (end & 0xFFF);
			size -= space->space;
		} else {
			space->space = size;
			size = 0;
		}
		
		addr = (addr & ~0xFFF) + 0x1000;
	}
	
	return (void *) virtual_from;
}

void *mmu_malloc(unsigned int pid, unsigned int size) {
	if (size == 0) return (void *) NULL;
	task_mmu_list *tsk_desc = getTaskMMUDescriptor(pid);
	
	task_alloc *palloc = tsk_desc->alloc;
	if (palloc != NULL) {
		//Me fijo si lo puedo poner antes de palloc
		if ((palloc->virtual_from - DINAMIC_MEMORY_START) >= size)
			return allocate_after(tsk_desc, NULL, DINAMIC_MEMORY_START, size);
	
		//Veo si hay algun lugar en el medio
		task_alloc *nalloc = palloc->next;
		while (nalloc != NULL) {
			uint pend = palloc->virtual_from + palloc->size;
			uint dist = nalloc->virtual_from - pend;
			if (dist >= size) {
				//Puedo meter la info en el medio
				return allocate_after(tsk_desc, palloc, pend, size);
			}
			
			palloc = nalloc;
			nalloc = nalloc->next;
		}
		
		//Tengo que ponerlo luego del ultimo item de la lista (palloc)
		return allocate_after(tsk_desc, palloc, palloc->virtual_from + palloc->size, size);
	}
	
	//No esta usando memoria, puedo ponerlo al inicio
	return allocate_after(tsk_desc, NULL, DINAMIC_MEMORY_START, size);
}

void mmu_free(unsigned int pid, void *address) {
	uint addr = (uint) address;
	if (addr == NULL) return;
	
	task_mmu_list *tsk_desc = getTaskMMUDescriptor(pid);
	
	task_alloc *palloc = NULL;
	task_alloc *nalloc = tsk_desc->alloc;
	if (nalloc != NULL) {
		//Busco alloc tal que alloc->next->virtual_from > addr
		//O sea, con alloc->virtual_from < addr
		while (nalloc->next != NULL && nalloc->next->virtual_from <= addr) {
			palloc = nalloc;
			nalloc = nalloc->next;
		}
		
		uint virtual_from = nalloc->virtual_from;
		uint size = nalloc->size;
		if (virtual_from <= addr && addr < virtual_from + size) {
			//Si addr esta en el rango del alloc actual, tengo que liberarlo todo
			//Otra alternativa seria que solo pueda liberar memoria si la direccion
			//que recibo coincide con virtual_from
			
			tsk_desc->space_used -= size;
			
			//Libero los recursos de alloc
			//Chequea si era el primero en la lista
			if (palloc == NULL) {
				tsk_desc->alloc = nalloc->next;
			} else {
				palloc->next = nalloc->next;
			}
			
			//Como no usa spaces, solo queda liberar el recurso
			release_resource(task_alloc, nalloc);
			
			uint page = virtual_from &~0xFFF;
			uint last_page = (virtual_from + size);
			uint addr = virtual_from;
			uint sub_size = 0x1000 - (addr &0xFFF);		//Lo que le saco al size con cada iteracion
			
			//Reestructuro la parte de la memoria fisica
			task_phys_stat *pphys = NULL;
			task_phys_stat *nphys = tsk_desc->pages;
			for (; page < last_page; page += 0x1000) {
				//Me quedo con nphys tal que phys->virtual_from >= page, o bien NULL
				while (nphys != NULL && nphys->virtual_address < page) {
					pphys = nphys;
					nphys = nphys->next;
				}
				
				//Si nphys es valido y mapea virtualmente a page, hay que hacer algo
				if (nphys != NULL && nphys->virtual_address == page) {
					uint phys_start = (addr & 0xFFF) + nphys->page;
					uint phys_end = phys_start + size;
			
					//Voy eliminando los space tales que el intervalo [mem_from, mem_from + space]
					//intersecte con el intervalo [phys_start, phys_end]
					task_alloc_desc *pspace = NULL;
					task_alloc_desc *nspace = nphys->space;

					while (nspace != NULL) {
						/*if ((space->mem_from <= phys_end) && (space->mem_from >= phys_start)) ||
							((phys_start <= space->mem_from) && (phys_start >= phys_start))*/
							
						if ((nspace->mem_from > phys_end) || (phys_start > (nspace->mem_from + nspace->space))) {
							//These are not the spaces you are looking for
							pspace = nspace;
							nspace = nspace->next;
						} else {
							//Intersecta, debe ser eliminado!
							task_alloc_desc *rspace = nspace;
							nspace = nspace->next;
							
							//Chequea si era el primero en la lista
							if (pspace == NULL) {
								nphys->space = nspace;
							} else {
								pspace->next = nspace;
							}
							
							//Libero el recurso
							release_resource(task_alloc_desc, rspace);
							
							//Si la pagina fisica no tiene mas detalle, se recicla
							if (nphys->space == NULL) {
								task_phys_stat *rphys = nphys;
								nphys = nphys->next;
								
								//Chequea si era la primera en la lista
								if (pphys == NULL) {
									tsk_desc->pages = nphys;
								} else {
									pphys->next = nphys;
								}

								recycleUserPage(rphys->page);
								release_resource(task_phys_stat, rphys);
								
								//Y se desmapea
								unmapear_pagina(page, rcr3());
							}
						}
					}
				}
				
				size -= sub_size;
				sub_size = 0x1000;	//En la siguiente iteracion le saco 4k al size
				addr = (addr & ~0xFFF) + 0x1000;
			}
		}
	}
}


void mmu_free_all(unsigned int pid) {
	task_mmu_list *tsk_desc = getTaskMMUDescriptor(pid);
	
	task_alloc *nalloc = tsk_desc->alloc;
	while (nalloc != NULL) {
		//Recorro la parte de la memoria fisica
		task_phys_stat *nphys = tsk_desc->pages;
		while (nphys != NULL) {
			task_alloc_desc *nspace = nphys->space;
			while (nspace != NULL) {
				//Debe ser eliminado!
				task_alloc_desc *rspace = nspace;
				nspace = nspace->next;
				release_resource(task_alloc_desc, rspace);
			}

			//La pagina fisica no tiene mas detalle, se recicla
			task_phys_stat *rphys = nphys;
			nphys = nphys->next;
			tsk_desc->pages = nphys;
			
			recycleUserPage(rphys->page);
			release_resource(task_phys_stat, rphys);
		}

		//Libero el recurso
		task_alloc *ralloc = nalloc;
		nalloc = nalloc->next;
		release_resource(task_alloc, ralloc);
	}

	//Libero el tsk_desc
	task_mmu_list *plst = &mmu_info;
	if (plst != tsk_desc) {
		while (plst->next != tsk_desc) 	
			plst = plst->next;
		plst->next = tsk_desc->next;
		release_resource(task_mmu_list, tsk_desc);
	}
}

//Hack para poder acceder desde la tarea idle a la info guardada en 
//los malloc que hizo el kernel.
void mapear_paginas_malloc_a_tarea_idle() {
	unsigned int idle_cr3 = 0x108000;

	task_mmu_list *mmu_ptr = &mmu_info;
	while (mmu_ptr != NULL) {
		if (mmu_ptr->pid == 0) {
			task_phys_stat *pages_ptr = mmu_ptr->pages;
			while (pages_ptr != NULL) {
				mapear_pagina(pages_ptr->virtual_address, idle_cr3, pages_ptr->page);
				pages_ptr = pages_ptr->next;
			}
			break;
		}
		mmu_ptr = mmu_ptr->next;
	}
}


//Imprime en pantalla la informacion que la mmu tiene de cada proceso que
//tiene memoria dinamica alojada
void debug_current_status(char *str) {
	task_mmu_list *mmu_ptr = &mmu_info;
	
	int wid = getScreenWidth();
	uint times = 4;
	uint space_left = (wid / 8) - times * ((wid / 8) / times);
	
	scr_printf("\n%C[MMU] DEBUG (%s)", YELLOW, str);
	while (mmu_ptr != NULL) {
		scr_printf("\n%C  [PID %d]%C usa %d bytes\n", WHITE, mmu_ptr->pid, 
			CYAN, mmu_ptr->space_used);
		
		scr_printf("%C    [VIRT_ALLOC]\n", WHITE);
		task_alloc *alloc_ptr = mmu_ptr->alloc;
		while (alloc_ptr != NULL) {
			scr_printf("%Cfrom:%C 0x%h -%C size:%C 0x%h\n", GRAY, CYAN, 
				alloc_ptr->virtual_from, GRAY, CYAN, alloc_ptr->size);
		
			//task_alloc_desc *space;
			alloc_ptr = alloc_ptr->next;
		}
		
		scr_printf("%C    [PHYS_ALLOC]\n", WHITE);
		task_phys_stat *pages_ptr = mmu_ptr->pages;
		uint printed = 0;
		while (pages_ptr != NULL) {
			scr_printf("%C0x%h %C->%C 0x%h%p", CYAN, 
				pages_ptr->virtual_address, GRAY, CYAN,
				pages_ptr->page, (wid / 8) / times);
			printed ++;
			if (printed == times) {
				printed = 0;
				scr_printf("%p", space_left);
			}
		
			//task_alloc_desc *space;
			pages_ptr = pages_ptr->next;
		}
		
		mmu_ptr = mmu_ptr->next;
	}
	debugged ++;
	//if (debugged == 4) asm("hlt");

	int c1 = 0;
	int c2 = 0;
	int c3 = 0;
	int c4 = 0;
	int c5 = 0;
	mmu_recycle_list *it1 = free_mmu_recycle_list;
	while (it1 != 0) { it1 = it1->next; c1 ++;}
	task_alloc_desc *it2 = free_task_alloc_desc;
	while (it2 != 0) { it2 = it2->next; c2 ++;}
	task_phys_stat *it3 = free_task_phys_stat;
	while (it3 != 0) { it3 = it3->next; c3 ++;}
	task_mmu_list *it4 = free_task_mmu_list;
	while (it4 != 0) { it4 = it4->next; c4 ++;}
	task_alloc *it5 = free_task_alloc;
	while (it5 != 0) { it5 = it5->next; c5 ++;}
	 
	scr_printf("\n|mmu_recycle_list| = %d -- ", c1);
	scr_printf("|task_alloc_desc| = %d -- ", c2);
	scr_printf("|task_phys_stat| = %d -- ", c3);
	scr_printf("|task_mmu_list| = %d -- ", c4);
	scr_printf("|task_alloc| = %d\n", c5);

	c1 = 0;
	it1 = recycled_kernel_pages;
	while (it1 != 0) { it1 = it1->next; c1 ++;}

	c2 = 0;
	it1 = recycled_user_pages;
	while (it1 != 0) { it1 = it1->next; c2 ++;}
	scr_printf("kernel_start = %h -- user start = %h -- recycled pages [kernel=%d, user=%d]\n", kernelPage, userPage, c1, c2);
}