#ifndef __MMU_H__
#define __MMU_H__

#include "../common/types.h"
#include "sched.h"

#define INICIO_PAGINAS_KERNEL 	0x00100000
#define INICIO_PAGINAS_USUARIO	0x00201000
#define TAMANO_PAGINA 			0x1000

#define DINAMIC_MEMORY_START	0x500000

typedef struct mmu_recycle_list mmu_recycle_list;
typedef struct task_alloc_desc task_alloc_desc;
typedef struct task_phys_stat task_phys_stat;
typedef struct task_mmu_list task_mmu_list;
typedef struct task_alloc task_alloc;

struct task_alloc_desc {
	task_alloc_desc *next;
	uint mem_from;
	uint space;
};

struct task_alloc {
	task_alloc *next;
	uint virtual_from;
	uint size;
};

struct task_phys_stat {
	task_phys_stat *next;
	uint page;
	uint virtual_address;
	task_alloc_desc *space;
};

struct task_mmu_list {
	task_mmu_list *next;
	int pid;
	uint space_used;
	task_alloc *alloc;
	task_phys_stat *pages;
};

struct mmu_recycle_list {
	mmu_recycle_list *next;
	uint page;
};


void inicializar_mmu();
unsigned int pagina_libre_kernel();
unsigned int pagina_libre_usuario();
void recycleUserPage(unsigned int page);
void recycleKernelPage(unsigned int page);

unsigned int inicializar_dir_usuario();
unsigned int direccion_fisica(unsigned int mvirt, unsigned int cr3);
unsigned int mapear_memoria_video(unsigned int cr3);
void unmapear_memoria_video(unsigned int cr3, unsigned int dir_data);
unsigned int obtener_entrada_directorio(unsigned int mvirt, unsigned int cr3);
void setear_entrada_directorio(unsigned int mvirt, unsigned int cr3, unsigned int data);
void mapear_pagina(unsigned int mvirt, unsigned int cr3, unsigned int mfis);
void mapear_4mb(unsigned int mvirt, unsigned int cr3, unsigned int mfis);
void unmapear_pagina(unsigned int mvirt, unsigned int cr3);
void unmapear_4mb(unsigned int mvirt, unsigned int cr3);
void recycleDirectoryMemory(unsigned int dir);

void *mmu_malloc(unsigned int pid, unsigned int size);
void mmu_free(unsigned int pid, void *address);
void *malloc(unsigned int size);
void free(void *address);
void mmu_free_all(unsigned int pid);

void *get_more_resources(int size);
void *allocate_after(task_mmu_list *tsk_desc, task_alloc *prev_alloc, 
	uint virtual_from, uint size);

void mapear_paginas_malloc_a_tarea_idle();

void debug_current_status(char *str);

#endif