#include "sched.h"
#include "tss.h"
#include "gdt.h"
#include "mmu.h"
#include "isr.h"
#include "i386.h"
#include "../drivers/keyboard.h"
#include "../common/types.h"
#include "../common/utils.h"


unsigned int next_pid;
unsigned int taskTime;
unsigned int taskWithKeyboardFocus;

unsigned int freeTasksInd[CANT_TAREAS + 1]; //+1 por la tarea idle
task_list freeTasksMem[CANT_TAREAS + 1];
unsigned int nextFreeTaskId;

task_list *actList;
task_list *slpList;
task_list *idleTaskEntry;
task_list *lastLoadXMMTaskEntry;
unsigned int activeTaskRemoved;


task_list *getFreeTaskEntry() {
	if (nextFreeTaskId < 0) return NULL;
	return &(freeTasksMem[freeTasksInd[nextFreeTaskId --]]);
}
void releaseTaskEntry(task_list *tsk) {
	tsk->prev = NULL;
	tsk->next = NULL;
	nextFreeTaskId ++;
	freeTasksInd[nextFreeTaskId] = tsk->recycleId;
}
void removeTaskEntryFromList(task_list *tsk) {
	tsk->next->prev = tsk->prev;
	tsk->prev->next = tsk->next;
	tsk->next = tsk;
	tsk->prev = tsk;
}
void addTaskEntryToList(task_list *tsk, task_list *list) {
	tsk->next = list;
	tsk->prev = list->prev;
	list->prev->next = tsk;
	list->prev = tsk;
}

unsigned short inicializar_scheduler(unsigned int idle_addr) {
	scr_printf("%C[KERNEL]%C Inicializando Scheduler\n", YELLOW, WHITE);

	int i = 0;
	for (; i <= CANT_TAREAS; i++) {
		freeTasksInd[i] = i;
		freeTasksMem[i].recycleId = i;
		freeTasksMem[i].prev = NULL;
		freeTasksMem[i].next = NULL;
	}
	nextFreeTaskId = CANT_TAREAS;

	act_pid = IDLE_PID;
	next_pid = IDLE_PID;

	taskTime = 0;
	actList = NULL;
	slpList = NULL;
	idleTaskEntry = NULL;
	activeTaskRemoved = 0;
	taskWithKeyboardFocus = -1;
	lastLoadXMMTaskEntry = NULL;

	//Inicializa la tarea IDLE
	int r = crear_proceso(idle_addr, 4);
	idleTaskEntry = actList;

	//Mapea las paginas que se hayan mapeado por un malloc
	//al directorio de la tarea idle
	mapear_paginas_malloc_a_tarea_idle();
	return r;
}

void addToSleepList(task_list *taskEntry) {
	if (slpList == NULL) {
		slpList = taskEntry;
		slpList->next = slpList;
		slpList->prev = slpList;
	} else {
		addTaskEntryToList(taskEntry, slpList);
	}
}
void sleep() {
	if (actList == idleTaskEntry) return;

	//La elimina de la lista de tareas activas
	activeTaskRemoved = 1;
	task_list *nextTask = actList->next;
	if (actList == actList->next) {
		//No hay mas tareas activas
		nextTask = idleTaskEntry;
	} else {
		removeTaskEntryFromList(actList);
	}

	//La agrega a la lista de tareas dormidas
	addToSleepList(actList);

	actList = nextTask;
	switch_task();
}
void setTaskActiveTime(unsigned int time) {
	actList->tsk.activeTime = time;
}
void sleepTask(unsigned int time) {
	if (time > 0) {
		actList->tsk.sleepTime = time;
		sleep();
	}
}

void waitTask(unsigned int time) {
	actList->tsk.waitTime = time;
	while (actList->tsk.waitTime > 0);
}

void loadXMM() {
	//En el caso en que sean iguales y lastLoadXMMTaskEntry haya sido reciclado
	//y reutilizado, no tiene sentido hacer fxsave
	if (lastLoadXMMTaskEntry != actList) {
		if (lastLoadXMMTaskEntry != NULL) {
			_fxsave((unsigned int) &(lastLoadXMMTaskEntry->tsk.fxData));
			lastLoadXMMTaskEntry->tsk.fxValid = 1;
		}
		
		if (actList->tsk.fxValid == 1) {
			//Carga el nuevo
			_fxrstor((unsigned int) &(actList->tsk.fxData));
		}
		lastLoadXMMTaskEntry = actList;
	}
}

void doStep() {
	taskTime ++;

	if (actList != idleTaskEntry && idleTaskEntry->tsk.waitTime > 0) 
		idleTaskEntry->tsk.waitTime --;

	//Cicla en las tareas activas
	task_list *it = actList;
	do {
		if (it->tsk.waitTime > 0) it->tsk.waitTime --;
		it = it->next;
	} while (it != actList);

	//Cicla en las tareas dormidas
	it = slpList;
	int cont = 0;
	if (it != NULL) {
		do {
			cont = 0;
			task_list *next = it->next;
			//if (it->tsk.status == 0 && it->tsk.sleepTime > 0) {
			if (it->tsk.sleepTime > 0) {
				it->tsk.sleepTime --;
				if (it->tsk.sleepTime == 0) {
					//La tiene que sacar de la lista de dormidas
					//y agregar a la lista de activas
					slpList = next;
					removeTaskEntryFromList(it);
					addTaskEntryToList(it, actList);
					
					if (slpList == it) {
						//Era la unica tarea durmiendo
						slpList = NULL;
						next = NULL;
					} else {
						cont = 1; //para evitar salir ahora que next == slpList
					}
				}
			}

			it = next;
		} while (cont || it != slpList);
	}
}

void releaseTaskSystemResources(task_list *tsk_entry) {
	taskStr *tsk = &(tsk_entry->tsk);

	//Libera las paginas de codigo
	int i = 0;
	for (; i < tsk->pages * 0x1000; i += 0x1000) {
		unsigned int code = direccion_fisica(0x400000 + i, tsk->dir);
		recycleUserPage(code);
	}

	//Libera la pagina de stack
	recycleUserPage(tsk->stack_page);

	//Libera la memoria del directorio de paginas
	recycleDirectoryMemory(tsk->dir);

	//Libera toda la memoria dinamica usada
	mmu_free_all(tsk->pid);

	//Libera los segmentos y entradas de tss
	releaseSegmentAndTSS(tsk->gdtOffset);

	//Libera la entrada
	releaseTaskEntry(tsk_entry);
}

void forced_switch_task() {
	taskTime = actList->tsk.activeTime;
	switch_task();
}
void switch_task_return() {
	//Privilegios de ser la tarea idle
	if (act_pid == IDLE_PID) {
		//Borra las tareas que quedaron sin eliminar
		task_list *it = slpList;
		int cont = 0;
		if (it != NULL) {
			do {
				cont = 0;
				task_list *next = it->next;
				if (it->tsk.delete_later == 1) {
					//Dispose of it.
					slpList = next;

					if (slpList == it) {
						//Era la unica tarea durmiendo
						slpList = NULL;
						next = NULL;
					} else {
						cont = 1; //para evitar salir dado que next == slpList
						removeTaskEntryFromList(it);
					}

					//scr_printf("deleting pid %d, gdt=%h, recycleId=%d | slpList (ptr)=%d\n", it->tsk.pid, it->tsk.gdtOffset, it->recycleId, slpList);

					//Libera todo recurso utilizado por la tarea
					releaseTaskSystemResources(it);
				}
				it = next;
			} while (cont || it != slpList);
		}
	}
}
unsigned short proximo_indice() {
	if (activeTaskRemoved) {
		activeTaskRemoved = 0;
		taskTime = 0;
	} else {
		if (actList != actList->next) {
			if (taskTime < actList->tsk.activeTime)
				return actList->tsk.gdtOffset;

			taskTime = 0;
			if (actList == idleTaskEntry) { //Si la actual es la tarea idle
				actList = actList->next;
				removeTaskEntryFromList(idleTaskEntry);
			} else {
				actList = actList->next;
			}
		}
	}

	act_pid = actList->tsk.pid;
	return actList->tsk.gdtOffset;
}

void s_getKeyboardFocus() {
	if (taskWithKeyboardFocus == actList->tsk.pid) return;
	taskWithKeyboardFocus = actList->tsk.pid;
	k_startBufferingCodes();
}
char s_getKeyboardNextCode() {
	if (taskWithKeyboardFocus != actList->tsk.pid) return 0;
	return k_getNextCode();
}
char s_getKeyboardNextChar() {
	if (taskWithKeyboardFocus != actList->tsk.pid) return 0;
	return k_getNextChar();
}
char s_getKeyboardLastCode() {
	if (taskWithKeyboardFocus != actList->tsk.pid) return 0;
	return k_getLastCode();
}
char s_getKeyboardLastChar() {
	if (taskWithKeyboardFocus != actList->tsk.pid) return 0;
	return k_getLastChar();
}

unsigned short crear_proceso(unsigned int cargar_desde, unsigned int pages) {
	task_list *tskEntry = getFreeTaskEntry();
	if (tskEntry == NULL) return NULL;

	unsigned int dir = inicializar_dir_usuario();
	unsigned int stack = pagina_libre_usuario();
	
	mapear_pagina(stack, dir, stack);
	unsigned short segment = nuevaTarea(dir, stack + 0x1000 - 24, 0x400000);
	
	unsigned int i = 0;
	for (i = 0; i < pages * 0x1000; i += 0x1000) {
		unsigned int code = pagina_libre_usuario();
		mapear_pagina(0x400000 + i, dir, code);
		
		//Guardo las direcciones de las paginas que voy a cambiar
		unsigned int dfis1 = direccion_fisica(code, rcr3());
		unsigned int dfis2 = direccion_fisica(code + 4095, rcr3());
		
		mapear_pagina(code, rcr3(), code);
		mapear_pagina(code + 4095, rcr3(), code + 4095);
		unsigned int *from = (unsigned int *) (cargar_desde + i);
		unsigned int *to = (unsigned int *) code;
		unsigned int j = 0;
		while (j < 1024) {
			to[j] = from[j];
			j ++;
		}
		
		//Remapeo las paginas
		if (dfis1 & 1) {
			mapear_pagina(code, rcr3(), dfis1);
		} else {
			unmapear_pagina(code, rcr3());
		}
		//Remapeo las paginas
		if (dfis2 & 1) {
			mapear_pagina(code + 4095, rcr3(), dfis2);
		} else {
			unmapear_pagina(code + 4095, rcr3());
		}
	}

	taskStr *tsk = &(tskEntry->tsk);
	tsk->keyCode = 0;
	tsk->pid = next_pid ++;
	tsk->gdtOffset = segment;
	tsk->delete_later = 0;
	tsk->waitTime = 0;
	tsk->sleepTime = 0;
	tsk->activeTime = 10;
	tsk->fxValid = 0;

	tsk->stack_page = stack;
	tsk->pages = pages;
	tsk->dir = dir;

	tsk->parent_pid = act_pid;
	
	//Si es la 1ra tarea creada (la idle), considera que esta tarea 
	//seria la unica activa
	if (actList == NULL) {
		actList = tskEntry;
		tskEntry->next = tskEntry;
		tskEntry->prev = tskEntry;
	} else {
		addTaskEntryToList(tskEntry, actList);
	}

	return segment;
}

void kill_current_task() {
	if (actList != idleTaskEntry) {
		//Tengo que marcarla para ser borrada despues
		//y concatenarla a la lista de 'sleep'
		task_list *currTask = actList;
		currTask->tsk.delete_later = 1;

		//Si era su propio siguiente, entonces no hay mas
		//tareas activas, y tiene que saltar a la idle
		actList = actList->next;
		if (actList == currTask) {
			actList = idleTaskEntry;
		} else {
			removeTaskEntryFromList(currTask);
		}
		activeTaskRemoved = 1;

		if (taskWithKeyboardFocus == currTask->tsk.pid) 
			taskWithKeyboardFocus = -1;

		addToSleepList(currTask);
		forced_switch_task();
	}
}

void s_specialKeyPressed() {
	//Si la tarea idle esta en la lista de activas, es la actual
	if (actList == idleTaskEntry) {
		//En ese caso, la saco
		actList = actList->next;
		removeTaskEntryFromList(idleTaskEntry);
	}
	if (actList != idleTaskEntry) {
		//La tarea idle no era la unica
		//Tengo que marcar a todas las tareas que quedan
		//para ser borradas despues
		//y concatenarlas a la lista de 'sleep'
		if (slpList == NULL) {
			slpList = actList;
		} else {
			task_list *aux = actList->prev;
			slpList->prev->next = actList;
			actList->prev->next = slpList;
			actList->prev = slpList->prev;
			slpList->prev = aux;
		}

		task_list *it = actList;
		do {
			it->tsk.delete_later = 1;
			it = it->next;
		} while (it != actList);
	}
	//*/

	if (taskWithKeyboardFocus != idleTaskEntry->tsk.pid) 
		taskWithKeyboardFocus = -1;

	actList = idleTaskEntry;
	actList->next = actList;
	actList->prev = actList;
	activeTaskRemoved = 1;
	forced_switch_task();
}