#include "tss.h"
#include "i386.h"
#include "mmu.h"

tss tsss[TSS_COUNT];
tss tarea_inicial;
unsigned int freeTsssIndex[TSS_COUNT];
unsigned int nextFreeTsssId;

void inicializar_tareas() {
	inicializar_gdt();

	int i = 0;
	for (; i < TSS_COUNT; i++) 
		freeTsssIndex[i] = i;
	nextFreeTsssId = TSS_COUNT - 1;
	
	ltr(nuevoSegmento(&tarea_inicial));
}

tss *nuevoTSS(unsigned int cr3, unsigned int esp, unsigned int eip) {
	if (nextFreeTsssId < 0) return NULL;
	tss *ret = &tsss[freeTsssIndex[nextFreeTsssId --]];

	inicializarTSS(ret, cr3, esp, eip);
	return ret;
}
unsigned short nuevaTarea(unsigned int cr3, unsigned int esp, unsigned int eip) {
	tss *newTss = nuevoTSS(cr3, esp, eip);
	return nuevoSegmento(newTss);
}
void inicializarTSS(tss* tssEntry, unsigned int cr3, unsigned int esp, unsigned int eip) {
	tssEntry->ptl = 0;
	tssEntry->esp0 = esp;
	tssEntry->esp = esp;
	tssEntry->ebp = 0;
	
	tssEntry->ss0 = 0x10;
	tssEntry->ss = 0x10;
	tssEntry->cs = 0x08;
	tssEntry->es = 0x10;
	tssEntry->fs = 0x10;
	tssEntry->gs = 0x10;
	tssEntry->ds = 0x10;
	
	tssEntry->eax = 0;
	tssEntry->ebx = 0;
	tssEntry->ecx = 0;
	tssEntry->edx = 0;
	tssEntry->esi = 0;
	tssEntry->edi = 0;
	tssEntry->eip = eip;
	
	tssEntry->cr3 = cr3;
	tssEntry->ldt = 0;
	tssEntry->eflags = 0x202;
	
	tssEntry->dtrap = 0;
	tssEntry->iomap = 0;
}

unsigned short nuevoSegmento(tss* tssEntry) {
	gdt_entry* gdtEntry = entradaLibreGDT();
	return inicializarSegmento(gdtEntry, tssEntry);
}
unsigned short inicializarSegmento(gdt_entry *gdtEntry, tss* tssEntry) {
	unsigned int tssPos = (unsigned int) tssEntry;
	gdtEntry->base_0_15 = (unsigned short) (tssPos);
	gdtEntry->base_23_16 = (unsigned char) (tssPos >> 16);
	gdtEntry->base_31_24 = (unsigned char) (tssPos >> 24);
	
	gdtEntry->limit_0_15 = 0x67;
	gdtEntry->limit_16_19 = 0x0;
	
	gdtEntry->type = 0x9;
	gdtEntry->s = 0;
	gdtEntry->db = 1;
	gdtEntry->p = 1;
	gdtEntry->l = 0;
	gdtEntry->avl = 0;
	gdtEntry->g = 0;
	gdtEntry->dpl = 0;
	
	return (unsigned short) (((unsigned int) gdtEntry) - ((unsigned int) GDT_DESC.gdt_addr));
}

void releaseSegmentAndTSS(unsigned short segment) {
	gdt_entry *gdtEntry = (gdt_entry *) 
						(((unsigned int) segment) + ((unsigned int) GDT_DESC.gdt_addr));
	unsigned int tssPos =
						((unsigned int) gdtEntry->base_0_15) + 
						(((unsigned int) gdtEntry->base_23_16) << 16) + 
						(((unsigned int) gdtEntry->base_31_24) << 24);

	nextFreeTsssId ++;
	unsigned int recycleId = (tssPos - ((unsigned int) tsss)) / sizeof(tss);
	freeTsssIndex[nextFreeTsssId] = recycleId;

	releaseSegment(gdtEntry);
}